import numpy as np
import os #remove before final
import time #not neccessary?
from .check import *
from .close import *
from .local_func_v7 import *
from ..collision.sphere_proximity import *
from ..collision.collision import *
from .add_bifurcation import *
from .sample_triad import *
from .triangle import * #might not need
from .basis import *
from scipy import interpolate
from scipy.spatial import KDTree
import matplotlib.pyplot as plt #remove before final
from .get_point import *
from mpl_toolkits.mplot3d import proj3d #remove before final
from .geodesic import extract_surface,geodesic
from ..implicit.visualize.visualize import show_mesh
from .finite_difference import finite_difference
from scipy.sparse.csgraph import shortest_path
from .add_geodesic_path import *
import pyvista as pv
import random
from scipy.interpolate import griddata
from time import perf_counter

def add_branch(tree,low,high,threshold_exponent=1.5,threshold_adjuster=0.75,
               all_max_attempts=40,max_attemps=10,sampling=20,max_skip=8,
               flow_ratio=None,radius_buffer=0,isforest=False,threshold=None,radius_scale=4,
               method='L-BFGS-B'):
    number_edges       = tree.parameters['edge_num']
    if threshold is None:
        threshold = ((tree.boundary.volume)**(1/3)/(number_edges**threshold_exponent))
    mu                 = tree.parameters['mu']
    lam                = tree.parameters['lambda']
    gamma              = tree.parameters['gamma']
    nu                 = tree.parameters['nu']
    Qterm              = tree.parameters['Qterm']
    Pperm              = tree.parameters['Pperm']
    Pterm              = tree.parameters['Pterm']
    new_branch_found   = False
    if tree.homogeneous:
        subb = None
        search_time = 0
        close_time = 0
        constraint_time = 0
        local_time = 0
        collision_time = 0
        time1 = 0
        time2 = 0
        time3 = 0
        time4 = 0
        add_time = 0
        total_time = 0
        start_total = time.time()
        start_rng = time.time()
        nonconvex_solve = True
        if len(tree.rng_points) == 0:
            #print('repick')
            rng_points,_ = tree.boundary.pick(size=len(tree.boundary.tet_verts),homogeneous=True,replacement=False)
            rng_points = rng_points.tolist()
        else:
            rng_points = tree.rng_points
        total_attempts = 0
        attempt        = 0
        while not new_branch_found:
            start = time.time()
            #print(threshold_distance)
            #total_attempts = 0
            #attempt        = 0
            for att in range(all_max_attempts):
                #print(attempt)
                start1 = time.time()
                #point,_ = tree.boundary.pick(homogeneous=True)
                #time1 += time.time() - start1
                if len(rng_points) == 1:
                    #print('repick_low')
                    rng_points,_ = tree.boundary.pick(size=len(tree.boundary.tet_verts),homogeneous=True,replacement=False)
                    rng_points = rng_points.tolist()
                point = np.array(rng_points.pop(0))
                if number_edges < 200:
                    vessel, line_distances = close_exact(tree.data,point)
                else:
                    vessel, line_distances = close(tree.data,point)
                time1 += time.time() - start1
                start2 = time.time()
                line_distances_below_threshold = sum(line_distances < threshold)
                minimum_line_distance = min(line_distances)
                time2 += time.time() - start2
                start3 = time.time()
                ################
                # check to see the attempt number
                ################
                if attempt < max_attemps:
                    attempt += 1
                    total_attempts += 1
                else:
                    if not tree.convex:
                        cell_list_outter = set([])
                        ptss = []
                        for i in range(tree.data.shape[0]):
                            first  = tree.data[i,0:3]
                            third  = tree.data[i,3:6]
                            second = (first+third)/2
                            ptss.append(second.tolist())
                            #cell_list_outter.update(set(tree.boundary.cell_lookup.query(first,k=tree.boundary.tet_verts.shape[0]//2)))
                            #cell_list_outter.update(set(tree.boundary.cell_lookup.query(third,k=tree.boundary.tet_verts.shape[0]//2)))
                            #cell_list_outter.update(set(tree.boundary.cell_lookup.query(second,k=tree.boundary.tet_verts.shape[0]//2)))
                            #cell_list_outter.update(set(tree.boundary.cell_lookup.query_ball_point(first,threshold+threshold*0.5)))
                            #cell_list_outter.update(set(tree.boundary.cell_lookup.query_ball_point(third,threshold+threshold*0.5)))
                            #d,id = tree.boundary.cell_lookup.query(second,5000)
                            #id = id.tolist()
                            #id = id[::-1]
                            #cell_list_outter.update(set(id[:200]))
                            cell_list_outter.update(set(tree.boundary.cell_lookup.query_ball_point(second,threshold)))
                            if threshold > ((tree.boundary.volume)**(1/3)/(number_edges**threshold_exponent))*(threshold_adjuster**5):
                                #pass
                                #cell_list_outter.difference_update(set(tree.boundary.cell_lookup.query_ball_point(first,threshold)))
                                #cell_list_outter.difference_update(set(tree.boundary.cell_lookup.query_ball_point(third,threshold)))
                                cell_list_outter.difference_update(set(tree.boundary.cell_lookup.query_ball_point(second,threshold*threshold_adjuster)))
                            ball_size = 2*threshold
                            while len(cell_list_outter) < 1:
                                cell_list_outter.update(set(tree.boundary.cell_lookup.query_ball_point(second,ball_size)))
                                ball_size *= 2
                        cell_list = list(cell_list_outter)
                        #plotter = pv.Plotter()
                        subb = tree.boundary.tet.grid.extract_cells(cell_list)
                        #plotter.add_mesh(tree.boundary.tet.grid,opacity=0.5)
                        #plotter.add_mesh(subb)
                        #point_poly = pv.PolyData(np.array(ptss))
                        #plotter.add_mesh(point_poly,color='red')
                        #plotter.show()
                        rng_points = []
                        max_attemps = min(100,len(cell_list))
                        for j in range(len(cell_list)):
                            p,_ = tree.boundary.pick_in_cell(cell_list[j])
                            rng_points.append(p[0].tolist())
                        random.shuffle(rng_points)
                    threshold *= threshold_adjuster
                    attempt = 0
                    #print(threshold)
                ##################
                ##################
                #print("Below Threshold: {}".format(line_distances_below_threshold))
                #print("Minimum Distances: {}  Threshold: {}".format(minimum_line_distance,radius_scale*tree.data[vessel[0],21]))
                if (line_distances_below_threshold == 0 and
                    minimum_line_distance > radius_scale*tree.data[vessel[0],21]):
                    escape = False
                    start3 = time.time()
                    for i in range(max_skip):
                        if flow_ratio is not None:
                            if tree.data[vessel[i],22] < flow_ratio*Qterm:
                                vessel = vessel[i]
                                escape = True
                                break

                        else:
                            vessel = vessel[i]
                            escape = True
                            break
                    time3 += time.time() - start3
                    if escape:
                        #print('viable')
                        break
                    """
                    if attempt < max_attemps:
                        attempt += 1
                        total_attempts += 1
                    else:
                        #print('adjusting threshold')
                        if not tree.convex:
                            p0 = (tree.data[vessel[i],0:3]+tree.data[vessel[i],3:6])/2
                            p1 = point
                            path,lengths,res,pf,f,rm = boundary.find_best_path(p0,p1,niter=10)
                            path = np.array(path)
                            data = np.vstack((t.data,np.zeros((path,tree.data.shape[1]))))
                            data,sub_div_ind,sub_div_map = add_geodesic_path(data,path,lengths,vessel[i],
                                                                             sub_division_index,sub_division_map)
                            return vessel[i],data,sub_div_ind,sub_div_map
                        threshold_distance *= threshold_adjuster
                        attempt = 0
                    """
                else:
                    #print('line distance')
                    start4 = time.time()
                    if attempt < max_attemps:
                        attempt += 1
                        total_attempts += 1
                    else:
                        #print('adjusting threshold')
                        attempt = 0
                        threshold *= threshold_adjuster
                        continue
                        #attempt = 0
                    time4 += time.time() - start4
            search_time += time.time()-start
            start = time.time()
            if not isinstance(vessel,np.int64):
                if len(vessel) > 1:
                    vessel = vessel[0]
            proximal = tree.data[vessel,0:3]
            distal   = tree.data[vessel,3:6]
            terminal = point
            #print(distal)
            #print(type(distal))
            #print(terminal)
            #print(type(terminal))
            if np.all(terminal.shape != distal.shape):
                terminal = terminal.flatten()
            points   = get_local_points(tree.data,vessel,terminal,sampling,tree.clamped_root)
            #high_res_points = get_local_points(tree.data,vessel,terminal,10*sampling,tree.clamped_root)
            #points   = np.array(relative_length_constraint(points,proximal,distal,terminal,0.25))
            #P = forest.show()
            #P.add_points(points,render_points_as_spheres=True,point_size=20)
            #P.add_points(terminal,render_points_as_spheres=True,point_size=20,color='g')
            #P.show()
            points   = np.array(relative_length_constraint(points,proximal,distal,terminal,0.25))
            #high_res_points = np.array(relative_length_constraint(high_res_points,proximal,distal,terminal,0.25))
            #P = forest.show()
            #P.add_points(points,render_points_as_spheres=True,point_size=20)
            #P.add_points(terminal,render_points_as_spheres=True,point_size=20,color='g')
            #P.show()
            if not tree.convex:
                points = boundary_constraint(points,tree.boundary,2)
                #high_res_points = boundary_constraint(high_res_points,tree.boundary,2)
            if len(points) == 0:
                attempt += 1
                #print('constraint 1')
                continue
            if vessel != 0 and not tree.clamped_root:
                points = np.array(angle_constraint(points,terminal,distal,-0.4,True))
                #high_res_points = np.array(angle_constraint(high_res_points,terminal,distal,-0.4,True))
            if len(points) == 0:
                attempt += 1
                #print('constraint 2')
                continue
            if vessel != 0 and not tree.clamped_root:
               points = np.array(angle_constraint(points,terminal,distal,0.75,False))
               #high_res_points = np.array(angle_constraint(high_res_points,terminal,distal,0.75,False))
            if len(points) == 0:
                attempt += 1
                #print('constraint 3')
                continue
            if vessel != 0 and not tree.clamped_root:
                points = np.array(angle_constraint(points,terminal,proximal,0,False))
                #high_res_points = np.array(angle_constraint(high_res_points,terminal,proximal,0.2,False))
            if len(points) == 0:
                attempt += 1
                #print('constraint 4')
                continue
            if vessel != 0 and not tree.clamped_root:
                points = np.array(angle_constraint(points,distal,proximal,0,False))
                #high_res_points = np.array(angle_constraint(high_res_points,distal,proximal,0.2,False))
            if len(points) == 0:
                attempt += 1
                #print('constraint 5')
                continue
            """
            if vessel != 0 and not tree.clamped_root:
                points = np.array(angle_constraint(points,distal,proximal,0.2,False))
                #high_res_points = np.array(angle_constraint(high_res_points,distal,proximal,0.2,False))
            if len(points) == 0:
                attempt += 1
                #print('constraint 5')
                continue
            if vessel != 0 and not tree.clamped_root:
                points = np.array(angle_constraint(points,distal,proximal,0.2,False))
                #high_res_points = np.array(angle_constraint(high_res_points,distal,proximal,0.2,False))
            if len(points) == 0:
                attempt += 1
                #print('constraint 5')
                continue
            """
            if tree.data[vessel,17] >= 0:
                p_vessel = int(tree.data[vessel,17])
                vector_1 = -tree.data[p_vessel,12:15]
                vector_2 = (points - proximal)/np.linalg.norm(points - proximal,axis=1).reshape(-1,1)
                #vector_3 = (high_res_points - proximal)/np.linalg.norm(high_res_points - proximal,axis=1).reshape(-1,1)
                angle = np.array([np.dot(vector_1,vector_2[i]) for i in range(len(vector_2))])
                #high_res_angle = np.array([np.dot(vector_1,vector_3[i]) for i in range(len(vector_3))])
                points = points[angle<0]
                #high_res_points = high_res_points[high_res_angle<0]
                if len(points) == 0:
                    attempt += 1
                    #print('constraint 6')
                    continue


            tmp_points = []
            #high_res_tmp_points = []
            if not tree.convex:
                for pt in range(points.shape[0]):
                    if tree.boundary.within(points[pt,0],points[pt,1],points[pt,2],2):
                        tmp_points.append(points[pt,:])
                points = np.array(tmp_points)
                #for pt in range(high_res_points.shape[0]):
                #    if tree.boundary.within(high_res_points[pt,0],high_res_points[pt,1],high_res_points[pt,2],2):
                #        high_res_tmp_points.append(high_res_points[pt,:])
                #high_res_points = np.array(high_res_tmp_points)
                if len(points) == 0:
                    attempt += 1
                    #print('constraint 7')
                    continue
            tmp_points = []
            subdivision = 5
            #plotter = pv.Plotter()
            #polys = pv.PolyData(points)
            #term_poly = pv.PolyData(terminal)
            #plotter.add_mesh(tree.boundary.tet.grid,opacity=0.25)
            #plotter.add_mesh(polys,color='red')
            #plotter.add_mesh(term_poly,color='green')
            #if subb is not None:
            #    plotter.add_mesh(subb,color='blue',opacity=0.25)
            #plotter.show()
            if not tree.convex:
                for pt in range(points.shape[0]):
                    include = True
                    for sub in range(1,2*subdivision):
                        mid_proximal = points[pt,:]*(sub/(2*subdivision))+proximal*(1-sub/(2*subdivision))
                        mid_distal = points[pt,:]*(sub/(2*subdivision))+distal*(1-sub/(2*subdivision))
                        mid_terminal = points[pt,:]*(sub/(2*subdivision))+terminal*(1-sub/(2*subdivision))
                        mid_proximal = mid_proximal.flatten()
                        mid_distal = mid_distal.flatten()
                        mid_terminal = mid_terminal.flatten()
                        if vessel != 0:
                            if not tree.boundary.DD[0]((mid_proximal[0],mid_proximal[1],mid_proximal[2],len(tree.boundary.patches)//10)) < 0.1:
                                #print('proximal')
                                include = False
                                break
                        val = tree.boundary.DD[0]((mid_distal[0],mid_distal[1],mid_distal[2],len(tree.boundary.patches)//10))
                        if not val < 0.01:
                            #print(val)
                            #plotter = pv.Plotter()
                            #plotter.add_mesh(tree.boundary.tet.grid,opacity=0.5)
                            #val_poly = pv.PolyData(mid_distal)
                            #plotter.add_mesh(sub)
                            #plotter.add_mesh(val_poly,color='yellow')
                            #plotter.show()
                            #print('distal')
                            include = False
                            break
                        if not tree.boundary.DD[0]((mid_terminal[0],mid_terminal[1],mid_terminal[2],len(tree.boundary.patches)//10)) < 0.01:
                            #print('terminal')
                            include = False
                            break
                    if include:
                        tmp_points.append(points[pt,:])
                points = np.array(tmp_points)
            #plotter.show()
            if len(points) == 0:
                attempt += 1
                nonconvex_solve=False
                #print('constraint 8')
                continue
            #print('passed all constraints')

            constraint_time += time.time()-start
            start = perf_counter()
            #construct_results = constructor(tree.data,terminal,
            #                      vessel,gamma,nu,Qterm,Pperm,Pterm,lam,mu,sampling,method=method)
            #print('Constructor Time: {}'.format(perf_counter()-start))
            construct_time=perf_counter()-start
            #print("Minimize x: {} Value: {}".format(results[5],np.pi*results[0]**lam*results[1]**mu))
            #start = perf_counter()
            #fd_point,fd_idx,fd_volume,fd_trial = finite_difference(tree.data,points,terminal,
            #                                                       vessel,gamma,nu,Qterm,Pperm,Pterm)
            #fd_time = perf_counter()
            #print('Finite Difference Time: {}'.format(perf_counter()-start))
            #print("Finite difference: {} Value: {}".format(fd_point,fd_volume[fd_idx]))
            #start = time.time()
            start = perf_counter()
            results = fast_local_function(tree.data,points,terminal,
                                          vessel,gamma,nu,Qterm,Pperm,Pterm)
            brute_time = perf_counter()-start
            #truth_results = fast_local_function(tree.data,high_res_points,terminal,
            #                                    vessel,gamma,nu,Qterm,Pperm,Pterm)
            #print('Brute Time: {}'.format(perf_counter()-start))
            local_time += time.time()-start
            volume  = np.pi*(results[0]**lam)*(results[1]**mu)
            brute = volume
            idx     = np.argmin(volume)
            bif     = results[5][idx]
            #truth_volume = np.pi*(truth_results[0]**lam)*(truth_results[1]**mu)
            #truth_idx = np.argmin(truth_volume)
            #truth_bif = truth_results[5][truth_idx]
            #truth_best_vol = truth_volume[truth_idx]
            #print("Brute x: {} Value: {}".format(bif,min(volume)))
            start = time.time()
            no_collision = collision_free(tree.data,results,idx,terminal,
                                          vessel,radius_buffer)
            collision_time += time.time()-start

            if no_collision:
                new_branch_found = True
                start = time.time()
                data,sub_division_map,sub_division_index = add_bifurcation(tree,vessel,terminal,
                                                                           results,idx,isforest=isforest)
                #brute = np.sum(np.pi*data[:,21]**lam*data[:,20]**mu)
                #print("Brute x: {} Value: {}".format(bif,brute[idx]))
                #fig = plt.figure()
                #ax1 = fig.add_subplot(121)
                #x = np.linspace(min(points[:,0]),max(points[:,0]),100)
                #y =  np.linspace(min(points[:,1]),max(points[:,1]),100)
                #X, Y = np.meshgrid(x,y)
                #fd_Ti = griddata((points[:,0],points[:,1]),fd_volume,(X,Y),method='linear')
                #ax1.contour(X,Y,fd_Ti,linewidths=0.5,colors='k')
                #ax1.pcolormesh(X,Y,fd_Ti,shading='auto',cmap=plt.get_cmap('rainbow'))
                #ax1.colorbar()
                #brute_Ti = griddata((points[:,0],points[:,1]),brute,(X,Y),method='linear')
                #ax2 = fig.add_subplot(122)
                #ax2.contour(X,Y,brute_Ti,linewidths=0.5,colors='k')
                #ax2.pcolormesh(X,Y,brute_Ti,shading='auto',cmap=plt.get_cmap('rainbow'))
                #ax2.colorbar()
                #plt.colorbar()
                #plt.show()
                add_time += time.time()-start
                total_time += time.time()-start_total
                tree.time['search'].append(search_time)
                tree.time['constraints'].append(constraint_time)
                tree.time['local_optimize'].append(local_time)
                tree.time['collision'].append(collision_time)
                tree.time['close_time'].append(close_time)
                tree.time['search_1'].append(time1)
                tree.time['search_2'].append(time2)
                tree.time['search_3'].append(time3)
                tree.time['search_4'].append(time4)
                tree.time['add_time'].append(add_time)
                tree.time['total'].append(total_time)
                #tree.time['brute_time'].append(brute_time)
                #tree.time['method_time'].append(construct_time)
                #tree.time['depth'].append(data[vessel,26])
                #tree.time['method_x_value'].append(construct_results[5].flatten())
                #tree.time['method_value'].append(np.pi*construct_results[0]**lam*construct_results[1]**mu)
                #tree.time['brute_x_value'].append(bif)
                #tree.time['brute_value'].append(brute[idx])
                #tree.time['truth_x_value'].append(truth_bif)
                #tree.time['truth_value'].append(truth_best_vol)
                if nonconvex_solve:
                    tree.nonconvex_counter += 1
                else:
                    tree.nonconvex_counter = 0
                if tree.nonconvex_counter > 100 or tree.convex:
                    tree.convex =True
                else:
                    tree.convex =False
                #print("returning")
                return vessel,data,sub_division_map,sub_division_index,threshold
            else:
                #print('collision')
                attempt += 1
                continue
    else:
        reduced_data  = tree.data[tree.data[:,-1]>-1]
        segment_data  = tree.data[tree.data[:,-1]==-1]
        vessel        = np.random.choice(list(range(reduced_data.shape[0])))
        vessel_path   = segment_data[segment_data[:,29].astype(int)==vessel]
        other_vessels = segment_data[segment_data[:,29].astype(int)!=vessel]
        if reduced_data.shape[0] > 1:
            other_KDTree = KDTree((other_vessels[:,0:3]+other_vessels[:,3:6])/2)
        else:
            other_KDTree = None
        mesh,pa,cp,cd = tree.boundary.mesh(vessel_path[1:,0:3],threshold,threshold//fraction,dive=0,others=other_KDTree)
        D,PR = shortest_path(graph,directed=False,method="D",return_predecessors=True)
        bif_idx = set(list(range(mesh.shape[0])))



import numpy as np
from .basis import *
from .triangle import *

def to_point(start,end,grad,fraction,hess=None):
    path = []
    start = start.flatten()
    end = end.flatten()
    start_norm = grad(*start)
    t1,t2,n = tangent_basis(start_norm,start)
    theta = np.linspace(0,2*np.pi,num=40)
    potential_next_steps = start + t1*np.sin(theta).reshape(-1,1) + t2*np.cos(theta).reshape(-1,1)
    optimal_step = np.argmin(np.linalg.norm(potential_next_steps-end,axis=1)).flatten()[0]
    path.append(potential_next_steps[optimal_step])
    dist = np.linalg.norm(path[-1]-end)
    while dist > fraction:
        tmp_norm = grad(*path[-1])
        t1,t2,n = tangent_basis(tmp_norm,path[-1].flatten())
        potential_next_steps = path[-1] + t1*np.sin(theta).reshape(-1,1) + t2*np.cos(theta).reshape(-1,1)
        optimal_step = np.argmin(np.linalg.norm(potential_next_steps-end,axis=1)).flatten()[0]
        if dist - np.linalg.norm(potential_next_steps[optimal_step] - end) < 0.5*fraction:
            #direct_point = path[-1] + ((end - path[-1])/np.linalg.norm(end - path[-1]))*fraction
            #path.append(direct_point)
            break 
        else:
            path.append(potential_next_steps[optimal_step])
        dist = np.linalg.norm(path[-1] - end)
        #else:
        #    path.append(level_func(*potential_next_steps[pt_idx].flatten(),p0_target).x)
    path.append(end)
    path = np.array(path)
    return path

def sample_area(segment_data,term,grad,fraction):
    sides = []
    for i in range(segment_data.shape[0]):
        sides.append(to_point(segment_data[i,:],term,grad,fraction))
    points = np.vstack(sides)
    return points
    """
    side1 = to_point(p0,p1,grad,fraction)
    side2 = to_point(p0,p2,grad,fraction)
    side3 = to_point(p1,p2,grad,fraction)
    mid1 = side1[round(side1.shape[0]/2),:]
    mid2 = side2[round(side2.shape[0]/2),:]
    mid3 = side3[round(side3.shape[0]/2),:]
    mline1 = to_point(mid1,p2,grad,fraction)
    mline2 = to_point(mid2,p0,grad,fraction)
    mline3 = to_point(mid3,p1,grad,fraction)
    points = np.vstack((side1,side2,side3,mline1,mline2,mline3))
    """
    #return points



rom="""if not terminating:
    rom_simulation = simulation.ROM()
    params = simulation.ROMParameters()
    model_params = params.ModelParameters()

    model_params.name = '{}'+'_'+'{}'
    model_params.inlet_face_names = [str(inlet)]
    model_params.outlet_face_names = [str(o) for o in outlets]
    model_params.centerlines_file_name = '{}' + os.sep +'centerlines.vtp'

    mesh_params = params.MeshParameters()

    fluid_params = params.FluidProperties()
    material = params.WallProperties.OlufsenMaterial()

    bcs = params.BoundaryConditions()
    bcs.add_velocities(face_name=str(inlet),file_name='{}'+os.sep+'{}')
    for face_name in outlets:
        bcs.add_resistance(face_name=str(face_name),resistance=((1/resistance[face_name])/total_res)*0+outlet_res_total)
    solution_params = params.Solution()
    solution_params.time_step = {}
    solution_params.num_time_steps = {}
    outdir = '{}'
    rom_simulation.write_input_file(model_order={},model=model_params,
                                    mesh=mesh_params,fluid=fluid_params,
                                    material=material,boundary_conditions=bcs,
                                    solution=solution_params,directory=outdir)
    if {} == 0:
        zeroD_folder = '{}'
        run_rom = open(zeroD_folder+os.sep+'run.py','w+')
        run_rom_file = ''
        path = sv.__path__[0].replace(os.sep+'sv','')
        run_rom_file += 'import os\\n'
        run_rom_file += 'import sys\\n'
        run_rom_file += "sys.path.append('{{}}')\\n".format(path)
        run_rom_file += 'from svZeroDSolver import svzerodsolver\\n'
        input_file = '{}'+os.sep+'solver_0d.in'
        if 'Windows' in platform.system():
            input_file = input_file.replace(os.sep,os.sep+os.sep)
        run_rom_file += "svzerodsolver.solver.set_up_and_run_0d_simulation('{{}}')\\n".format(input_file)
        run_rom.writelines([run_rom_file])
        run_rom.close()

        rom_view = open(zeroD_folder+os.sep+'view.py','w+')
        rom_view_file = ''
        rom_view_file += 'import numpy as np\\n'
        rom_view_file += 'import matplotlib.pyplot as plt\\n'
        rom_view_file += 'data = np.load("solver_0d_branch_results.npy",allow_pickle=True).item()\\n'
        rom_view_file += 'vessels = []\\n'
        rom_view_file += 'fig_flow = plt.figure()\\n'
        rom_view_file += 'ax_flow = fig_flow.add_subplot()\\n'
        rom_view_file += 'fig_pressure = plt.figure()\\n'
        rom_view_file += 'ax_pressure = fig_pressure.add_subplot()\\n'
        rom_view_file += 'fig_flow_outlets = plt.figure()\\n'
        rom_view_file += 'ax_flow_outlets = fig_flow_outlets.add_subplot()\\n'
        rom_view_file += 'fig_pressure_outlets = plt.figure()\\n'
        rom_view_file += 'ax_pressure_outlets = fig_pressure_outlets.add_subplot()\\n'
        rom_view_file += 'for vessel in data["flow"]:\\n'
        rom_view_file += '    vessels.append(vessel)\\n'
        rom_view_file += 'for vessel in vessels:\\n'
        rom_view_file += '    ax_flow.plot(data["time"],data["flow"][vessel][0],label="vessel_"+str(vessel))\\n'
        rom_view_file += '    ax_pressure.plot(data["time"],data["pressure"][vessel][0]/1333.22,label="vessel_"+str(vessel))\\n'
        rom_view_file += 'import json\\n'
        rom_view_file += 'info = json.load(open("solver_0d.in"))\\n'
        rom_view_file += 'all_inlets = []\\n'
        rom_view_file += 'all_outlets = []\\n'
        rom_view_file += 'for i in info["junctions"]:\\n'
        rom_view_file += '    for j in i["inlet_vessels"]:\\n'
        rom_view_file += '        all_inlets.append(j)\\n'
        rom_view_file += '    for j in i["outlet_vessels"]:\\n'
        rom_view_file += '        all_outlets.append(j)\\n'
        rom_view_file += 'all_inlets = set(all_inlets)\\n'
        rom_view_file += 'all_outlets = set(all_outlets)\\n'
        rom_view_file += 'true_outlets = list(all_outlets.difference(all_inlets))\\n'
        rom_view_file += 'true_inlets = list(all_inlets.difference(all_outlets))\\n'
        rom_view_file += 'for vessel in true_outlets:\\n'
        rom_view_file += '    ax_flow_outlets.plot(data["time"],data["flow"][vessel][-1],label="vessel_"+str(vessel))\\n'
        rom_view_file += '    ax_pressure_outlets.plot(data["time"],data["pressure"][vessel][-1]/1333.22,label="vessel_"+str(vessel))\\n'
        rom_view_file += 'ax_flow.set_xlabel("Time (sec)")\\n'
        rom_view_file += 'ax_flow.set_ylabel("Flow (mL/s)")\\n'
        rom_view_file += 'ax_pressure.set_xlabel("Time (sec)")\\n'
        rom_view_file += 'ax_pressure.set_ylabel("Pressure (mmHg)")\\n'
        rom_view_file += 'ax_flow_outlets.set_xlabel("Time (sec)")\\n'
        rom_view_file += 'ax_pressure_outlets.set_xlabel("Time (sec)")\\n'
        rom_view_file += 'ax_flow_outlets.set_ylabel("Flow (mL/s)")\\n'
        rom_view_file += 'ax_flow_outlets.set_title("Outlets Only")\\n'
        rom_view_file += 'ax_pressure_outlets.set_ylabel("Pressure (mmHg)")\\n'
        rom_view_file += 'ax_pressure_outlets.set_title("Outlets Only")\\n'
        rom_view_file += 'plt.show()\\n'
        rom_view.writelines([rom_view_file])
        rom_view.close()
    if {} == 1:
        oneD_folder = '{}'
        print('Locating OneDSovler executable...')
        import sv
        from glob import glob
        import shutil
        if platform.system() == 'Windows':
            #search_path = sv.__path__[0].split('Python3.5')[0]
            search_path = 'C:\\\\Program Files\\\\SimVascular'
            files = glob(search_path+'\\\\**\\\\svOneDSolver.exe',recursive=True)
            if len(files) > 0:
                isfile = [os.path.isfile(f) for f in files]
                only_files = [files[of] for of in isfile if of]
                print('OneDSolver Found!')
                shutil.copy(only_files[0],oneD_folder)
            else:
                print('OneDSolver not found (will require user-specified executable location)')
        elif platform.system() == 'Linux':
            if 'simvascular' not in sv.__path__[0]:
                print('WARNING: Searching non-release build. OneDSolver not guarunteed to be installed.')
                print('Searching parent directory to simvascular build')
                search_dir = '/usr/local/sv'
                files = glob(search_dir+'/**/OneDSolver',recursive=True)
                if len(files) > 0:
                    isfile = [os.path.isfile(f) for f in files]
                    only_files = [files[of] for of in isfile if of]
                    print('OneDSolver Found!')
                    shutil.copy(only_files[0],oneD_folder)
                else:
                    print('OneDSolver not found (will require user-specified executable location)')
            else:
                #search_dir = sv.__path__[0].split('simvascular')[0]
                search_dir = '/usr/local/sv'
                files = glob(search_dir+'/**/OneDSolver',recursive=True)
                if len(files) > 0:
                    isfile = [os.path.isfile(f) for f in files]
                    only_files = [files[of] for of in isfile if of]
                    print('OneDSolver Found!')
                    shutil.copy(only_files[0],oneD_folder)
                else:
                    print('OneDSolver not found (will require user-specified executable location)')
        elif platform.system() == 'Darwin':
            # We will just assume the search directory here
            search_dir = '/usr/local/sv'
            files = glob(search_dir+'/**/OneDSolver',recursive=True)
            if len(files) > 0:
                isfile = [os.path.isfile(f) for f in files]
                only_files = [files[of] for of in isfile if of]
                print('OneDSolver Found!')
                shutil.copy(only_files[0],oneD_folder)
            else:
                print('OneDSolver not found (will require user-specified executable location)')
"""



import numpy as np
import os

#heart_points = np.genfromtxt(os.path.dirname(__file__)+'heart_points_unique.csv',delimiter=',')
#heart_normals = np.genfromtxt(os.path.dirname(__file__)+'heart_normals_unique.csv',delimiter=',')



from scipy.interpolate import splprep, splev
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
from mpl_toolkits.mplot3d import Axes3D, proj3d, art3d
from mpl_toolkits.mplot3d.art3d import Line3DCollection
import vtk
from scipy import optimize
from tqdm import tqdm
import pyvista as pv

from ..branch_addition.basis import tangent_basis

from mpl_toolkits.mplot3d import art3d

def rotation_matrix(d):
    """
    Calculates a rotation matrix given a vector d. The direction of d
    corresponds to the rotation axis. The length of d corresponds to
    the sin of the angle of rotation.

    Variant of: http://mail.scipy.org/pipermail/numpy-discussion/2009-March/040806.html
    """
    sin_angle = np.linalg.norm(d)

    if sin_angle == 0:
        return np.identity(3)

    d /= sin_angle

    eye = np.eye(3)
    ddt = np.outer(d, d)
    skew = np.array([[    0,  d[2],  -d[1]],
                  [-d[2],     0,  d[0]],
                  [d[1], -d[0],    0]], dtype=np.float64)

    M = ddt + np.sqrt(1 - sin_angle**2) * (eye - ddt) + sin_angle * skew
    return M

def pathpatch_2d_to_3d(pathpatch, z = 0, normal = 'z'):
    """
    Transforms a 2D Patch to a 3D patch using the given normal vector.

    The patch is projected into they XY plane, rotated about the origin
    and finally translated by z.
    """
    if type(normal) is str: #Translate strings to normal vectors
        index = "xyz".index(normal)
        normal = np.roll((1.0,0,0), index)

    normal /= np.linalg.norm(normal) #Make sure the vector is normalised

    path = pathpatch.get_path() #Get the path and the associated transform
    trans = pathpatch.get_patch_transform()

    path = trans.transform_path(path) #Apply the transform

    pathpatch.__class__ = art3d.PathPatch3D #Change the class
    pathpatch._code3d = path.codes #Copy the codes
    pathpatch._facecolor3d = pathpatch.get_facecolor #Get the face color

    verts = path.vertices #Get the vertices in 2D

    d = np.cross(normal, (0, 0, 1)) #Obtain the rotation vector
    M = rotation_matrix(d) #Get the rotation matrix

    pathpatch._segment3d = np.array([np.dot(M, (x, y, 0)) + (0, 0, z) for x, y in verts])

def pathpatch_translate(pathpatch, delta):
    """
    Translates the 3D pathpatch by the amount delta.
    """
    pathpatch._segment3d += delta

def get_longest_path(data,seed_edge):
    dig = True
    temp_edges = [seed_edge]
    while dig:
        keep_digging = []
        for edge in temp_edges:
            if data[edge,15] > -1:
                temp_edges.extend([int(data[edge,15])])
                if data[edge,16] > -1:
                    temp_edges.extend([int(data[edge,16])])
                temp_edges.remove(edge)
                keep_digging.append(True)
            else:
                keep_digging.append(False)
        dig = any(keep_digging)
    if len(temp_edges) == 1:
        return temp_edges
    edge_depths = []
    for edge in temp_edges:
        edge_depths.append(data[edge,26])
    max_depth = max(edge_depths)
    max_edge_depths = [i for i,j in enumerate(edge_depths) if j == max_depth]
    paths = [[temp_edges[i]] for i in max_edge_depths]
    path_lengths = [data[i[0],20] for i in paths]
    retrace = [True]*len(paths)
    while np.any(retrace):
        for i, path in enumerate(paths):
            if retrace[i]:
                path.insert(0,int(data[path[0],17]))
                path_lengths[i] += data[path[0],20]
                if paths[i][0] == seed_edge:
                    retrace[i] = False
    return paths[path_lengths.index(max(path_lengths))]

def get_alternate_path(data,seed_edge,reference=None):
    #print('reference: {}'.format(reference))
    if reference is None:
        reference = get_longest_path(data,seed_edge)
    else:
        pass
    seed_edge_idx = reference.index(seed_edge)
    if seed_edge_idx == len(reference)-1:
        print('Seed edge is terminal and no alternative is '+
              'possible. \n Computation finished.')
        return None
    else:
        children = [int(data[seed_edge,15]),int(data[seed_edge,16])]
        if children[0] not in reference:
            alternate_path = get_longest_path(data,children[0])
        else:
            alternate_path = get_longest_path(data,children[1])
    alternate_path.insert(0,seed_edge)
    return alternate_path

def get_branches(data):
    branches = []
    seed_edge = 0
    path = get_longest_path(data,seed_edge)
    branches.append(path)
    upcoming_evaluations = []
    upcoming_evaluations.extend(path[:-1])
    counter = [len(path[:-1])]
    idx = 0
    while len(upcoming_evaluations) > 0:
        #print(upcoming_evaluations)
        if data[upcoming_evaluations[-1],15] > -1 and data[upcoming_evaluations[-1],16] > -1:
            pass
        else:
            upcoming_evaluations.pop(-1)
            counter[idx] -= 1
            if counter[idx] == 0:
                counter[idx] = None
                for i in reversed(range(len(counter))):
                    if counter[i] is not None:
                        idx = i
                        break
            continue
        path = get_alternate_path(data,upcoming_evaluations.pop(-1),reference=branches[idx])
        counter[idx] -= 1
        if counter[idx] == 0:
            counter[idx] = None
            for i in reversed(range(len(counter))):
                if counter[i] is not None:
                    idx = i
                    break
        branches.append(path)
        if len(path) > 2:
            upcoming_evaluations.extend(path[1:-1])
            counter.append(len(path[1:-1]))
            idx = len(counter) - 1
        else:
            counter.append(None)
    return branches

def get_points(data,branches):
    path_points = []
    primed = False
    for path in branches:
        branch_points = []
        for edge in reversed(path):
            if edge == path[0] and primed:
                branch_points.insert(0,data[edge,3:6].tolist())
            elif edge == path[0] and edge == 0 and not primed:
                branch_points.insert(0,data[edge,3:6].tolist())
                three_fourths = data[edge,0:3]*(1/4) + data[edge,3:6]*(3/4) #new
                branch_points.insert(0,three_fourths.tolist()) #new
                mid_point = (data[edge,0:3] + data[edge,3:6])/2
                branch_points.insert(0,mid_point.tolist())
                one_fourth = data[edge,0:3]*(3/4) + data[edge,3:6]*(1/4) #new
                branch_points.insert(0,one_fourth.tolist()) #new
                branch_points.insert(0,data[edge,0:3].tolist())
                primed = True
            elif len(branch_points) == 0:
                branch_points.insert(0,data[edge,3:6].tolist())
                three_fourths = data[edge,0:3]*(1/4) + data[edge,3:6]*(3/4) #new
                branch_points.insert(0,three_fourths.tolist()) #new
                mid_point = (data[edge,0:3] + data[edge,3:6])/2
                branch_points.insert(0,mid_point.tolist())
                one_fourth = data[edge,0:3]*(3/4) + data[edge,3:6]*(1/4) #new
                branch_points.insert(0,one_fourth.tolist()) #new
            else:
                branch_points.insert(0,data[edge,3:6].tolist())
                three_fourths = data[edge,0:3]*(1/4) + data[edge,3:6]*(3/4) #new
                branch_points.insert(0,three_fourths.tolist()) #new
                mid_point = (data[edge,0:3] + data[edge,3:6])/2
                branch_points.insert(0,mid_point.tolist())
                one_fourth = data[edge,0:3]*(3/4) + data[edge,3:6]*(1/4) #new
                branch_points.insert(0,one_fourth.tolist()) #new
        path_points.append(branch_points)
    return path_points

def get_points_v1(data,branches):
    path_points = []
    primed = False
    for path in branches:
        branch_points = []
        for edge in reversed(path):
            if edge == path[-1]:
                branch_points.insert(0,data[edge,3:6].tolist())
            elif edge == path[0] and edge == 0 and not primed:
                branch_points.insert(0,data[edge,3:6].tolist())
                branch_points.insert(0,data[edge,0:3].tolist())
                primed = True
            else:
                branch_points.insert(0,data[edge,3:6].tolist())
        path_points.append(branch_points)
    return path_points

def get_radii(data,branches):
    path_radii = []
    primed = False
    for path in branches:
        branch_radii = []
        for edge in reversed(path):
            if edge == path[0] and primed:
                branch_radii.insert(0,data[path[1],21])
            elif edge == path[0] and edge == 0 and not primed:
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21]) #new
                branch_radii.insert(0,data[edge,21]) #new
                primed = True
            elif len(branch_radii) == 0:
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21]) #new
                branch_radii.insert(0,data[edge,21]) #new
            else:
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21])
                branch_radii.insert(0,data[edge,21]) #new
                branch_radii.insert(0,data[edge,21]) #new
        path_radii.append(branch_radii)
    return path_radii

def get_radii_v1(data,branches):
    path_radii = []
    primed = False
    for path in branches:
        branch_radii = []
        for edge in reversed(path):
            if edge == path[-1]:
                branch_radii.insert(0,data[edge,21])
            elif edge == path[0] and edge == 0 and not primed:
                branch_radii.insert(0,(data[edge,21]+data[previous,21])/2)
                branch_radii.insert(0,data[edge,21])
                primed = True
            else:
                branch_radii.insert(0,(data[edge,21]+data[previous,21])/2)
            previous = edge
        path_radii.append(branch_radii)
    return path_radii

def get_normals(data,branches):
    path_normals = []
    primed = False
    for path in branches:
        branch_normals = []
        for edge in reversed(path):
            if edge == path[0] and primed:
                branch_normals.insert(0,data[path[1],12:15].tolist())
            elif edge == path[0] and edge == 0 and not primed:
                vector_1 = data[edge,12:15]
                mid_vector = (vector_1+vector_2)/2
                branch_normals.insert(0,mid_vector.tolist())
                branch_normals.insert(0,data[edge,12:15].tolist())
                branch_normals.insert(0,data[edge,12:15].tolist())
                branch_normals.insert(0,data[edge,12:15].tolist()) #new
                branch_normals.insert(0,data[edge,12:15].tolist()) #new
                primed = True
            elif len(branch_normals) == 0:
                branch_normals.insert(0,data[edge,12:15].tolist())
                branch_normals.insert(0,data[edge,12:15].tolist())
                branch_normals.insert(0,data[edge,12:15].tolist()) #new
                branch_normals.insert(0,data[edge,12:15].tolist()) #new
                vector_2 = data[edge,12:15]
            else:
                vector_1 = data[edge,12:15]
                mid_vector = (vector_1+vector_2)/2
                branch_normals.insert(0,mid_vector.tolist())
                branch_normals.insert(0,data[edge,12:15].tolist()) #new
                branch_normals.insert(0,data[edge,12:15].tolist()) #new
                branch_normals.insert(0,data[edge,12:15].tolist())
                vector_2 = data[edge,12:15]
        path_normals.append(branch_normals)
    return path_normals

def get_normals_v1(data,branches):
    path_normals = []
    primed = False
    for path in branches:
        branch_normals = []
        for edge in reversed(path):
            if edge == path[-1]:
                branch_normals.insert(0,data[edge,12:15].tolist())
            elif edge == path[0] and edge == 0 and not primed:
                branch_normals.insert(0,((data[edge,12:15]+data[previous,12:15])/2).tolist())
                branch_normals.insert(0,data[edge,12:15].tolist())
                primed = True
            else:
                branch_normals.insert(0,((data[edge,12:15]+data[previous,12:15])/2).tolist())
            previous = edge
        path_normals.append(branch_normals)
    return path_normals

def plot_sv_data(data):
    branches = get_branches(data)
    points   = get_points(data,branches)
    radii    = get_radii(data,branches)
    normals  = get_normals(data,branches)
    fig = plt.figure()
    ax = fig.add_subplot(111,projection='3d')
    for path in points:
        tmp = np.array(path)
        ax.plot3D(tmp[:,0],tmp[:,1],tmp[:,2])
        ax.scatter3D(tmp[:,0],tmp[:,1],tmp[:,2])
    plt.show()

def get_interpolated_sv_data(data):
    branches = get_branches(data)
    points   = get_points(data,branches)
    #print("Points: {}".format(points[0]))
    radii    = get_radii(data,branches)
    #print("Radii:  {}".format(radii[0]))
    normals  = get_normals(data,branches)
    path_frames = []
    for idx in range(len(branches)):
        frames = []
        for jdx in range(len(points[idx])):
            frame = []
            frame.extend(points[idx][jdx])
            frame.append(radii[idx][jdx])
            frame.extend(normals[idx][jdx])
            frames.append(frame)
        path_frames.append(frames)
    interp_xyz = []
    interp_r   = []
    interp_n   = []
    interp_xyzr = []
    for idx in range(len(branches)):
        p = np.array(points[idx]).T
        r = np.array(radii[idx]).T
        n = np.array(normals[idx]).T
        if len(points[idx]) == 2:
            interp_xyz.append(splprep(p,k=1,s=0))
            rr = np.vstack((interp_xyz[-1][1],r))
            interp_r.append(splprep(rr,k=1,s=0))
            xyzr = np.vstack((p,r))
            interp_xyzr.append(splprep(xyzr,k=1,s=0))
            #interp_n.append(splprep(n,k=1,s=0))
        elif len(points[idx]) == 3:
            interp_xyz.append(splprep(p,k=2,s=0))
            rr = np.vstack((interp_xyz[-1][1],r))
            interp_r.append(splprep(rr,k=2,s=0))
            xyzr = np.vstack((p,r))
            interp_xyzr.append(splprep(xyzr,k=2,s=0))
            #interp_n.append(splprep(n,k=2,s=0))
        else:
            interp_xyz.append(splprep(p,s=0))
            rr = np.vstack((interp_xyz[-1][1],r))
            interp_r.append(splprep(rr,s=0))
            xyzr = np.vstack((p,r))
            interp_xyzr.append(splprep(xyzr,s=0))
            #interp_n.append(splprep(n,s=0))

    return interp_xyz,interp_r,interp_n,path_frames,branches,interp_xyzr

def get_sv_data(network):
    if isinstance(network,tree):
        data     = network.data
        branches = get_branches(data)
        points   = get_points(data,branches)
        radii    = get_radii(data,branches)
        normals  = get_normals(data,branches)
    elif isinstance(network,forest):
        for idx, grafts in enumerate(network.grafts):
            data        = []
            branches    = []
            connections = []
            for jdx, graft in enumerate(grafts):
                data.append(graft.data)
                branches.append(get_branches(graft.data))
                connections.append(network.connections[idx][jdx])

def plot_interp_vessels(interp_xyz,interp_r,normals=True):
    n = np.linspace(0,1,200)
    fig = plt.figure()
    ax1  = fig.add_subplot(111,projection='3d')
    #ax2  = fig.add_subplot(232,projection='3d')
    #ax3  = fig.add_subplot(233,projection='3d')
    #ax4  = fig.add_subplot(234,projection='3d')
    #ax5  = fig.add_subplot(235,projection='3d')
    #ax6  = fig.add_subplot(236,projection='3d')
    for idx in enumerate(range(interp_xyz)):
        if idx > 0:
            break
        x,y,z = splev(n,interp_xyz[idx][0])
        nx,ny,nz = splev(n,interp_xyz[idx][0],der=1)
        nxx,nyy,nzz = splev(n,inerp_xyz[idx][0],der=2)
        #nxxx,nyyy,nzzz,_,_,_,_ = splev(n,b[0],der=3)
        ax1.plot3D(x,y,z,c='b')
        ax1.scatter3D(x[0],y[0],z[0],c='black')
        ax1.scatter3D(x[-1],y[-1],z[-1],c='black')
        #ax2.plot3D(x,y,z,c='b')
        #ax3.plot3D(x,y,z,c='b')
        idx_x = np.argwhere(np.diff(np.sign(nx))!=0).flatten()
        idx_y = np.argwhere(np.diff(np.sign(ny))!=0).flatten()
        idx_z = np.argwhere(np.diff(np.sign(nz))!=0).flatten()
        ax1.scatter3D(x[idx_x],y[idx_x],z[idx_x],c='r')
        ax1.scatter3D(x[idx_y],y[idx_y],z[idx_y],c='r')
        ax1.scatter3D(x[idx_z],y[idx_z],z[idx_z],c='r')
        idx_xx = np.argwhere(np.diff(np.sign(nxx))!=0).flatten()
        idx_yy = np.argwhere(np.diff(np.sign(nyy))!=0).flatten()
        idx_zz = np.argwhere(np.diff(np.sign(nzz))!=0).flatten()
        ax1.scatter3D(x[idx_xx],y[idx_xx],z[idx_xx],c='g')
        ax1.scatter3D(x[idx_yy],y[idx_yy],z[idx_yy],c='g')
        ax1.scatter3D(x[idx_zz],y[idx_zz],z[idx_zz],c='g')
        if b[0][-1] >= 3:
            nxxx,nyyy,nzzz = splev(n,interp_xyz[idx][0],der=3)
            idx_xxx = np.argwhere(np.diff(np.sign(nxxx))!=0).flatten()
            idx_yyy = np.argwhere(np.diff(np.sign(nyyy))!=0).flatten()
            idx_zzz = np.argwhere(np.diff(np.sign(nzzz))!=0).flatten()
            ax1.scatter3D(x[idx_xxx],y[idx_xxx],z[idx_xxx],c='y')
            ax1.scatter3D(x[idx_yyy],y[idx_yyy],z[idx_yyy],c='y')
            ax1.scatter3D(x[idx_zzz],y[idx_zzz],z[idx_zzz],c='y')
        #ax.plot3D(x,y,z)
        #x_knot,y_knot,z_knot,_,_,_,_ = splev(b[1],b[0])
        #ax.plot3D(x_knot,y_knot,z_knot,'go')
        """
        if normals:
            nx = -1/nx
            ny = -1/ny
            nz = -1/nz
            l  = np.linalg.norm(np.array([nx,ny,nz]),axis=0)
            nx = (nx/l)*r
            ny = (ny/l)*r
            nz = (nz/l)*r
            points0 = np.array([x,y,z]).T.reshape(-1,1,3)
            points1 = np.array([x+nx,y+ny,z+nz]).T.reshape(-1,1,3)
            segments = np.concatenate([points0,points1],axis=1)
            lc = Line3DCollection(segments,color='black')
            ax.add_collection3d(lc)
        """
    plt.show()

from sympy import Point3D,Line3D,Plane,Float
def contour_collision(x1,y1,z1,nx1,ny1,nz1,r1,x2,y2,z2,nx2,ny2,nz2,r2,radius_buffer,contours,t1,t2,n_sides=4):
    """
    d1 = x1*nx1+y1*ny1+z1*nz1
    d2 = x2*nx2+y2*ny2+z2*nz2
    length = ((x2-x1)**2+(y2-y1)**2+(z2-z1)**2)**(1/2)
    if radius_buffer is not None:
        r1 += radius_buffer*r1
        r2 += radius_buffer*r2
    if not np.isclose(nx1,0):
        alpha = (nz1*nx2-nz2*nx1)/(ny2*nx1-ny1*nx2)
        beta  = (d1*nx2-d2*nx1)/(ny2*nx1-ny1*nx2)
        c     = ny1/nx1
        a_pt  = np.array([-d1/nx1-beta*c,beta,0])
        pt1   = np.array([x1,y1,z1])
        pt2   = np.array([x2,y2,z2])
        n_line= np.array([-(alpha*c+nz1/nx1),alpha,1])
        n_line=n_line/np.linalg.norm(n_line)
        dist1 = np.linalg.norm((pt1-a_pt)-np.dot(pt1-a_pt,n_line)*n_line)
        dist2 = np.linalg.norm((pt2-a_pt)-np.dot(pt2-a_pt,n_line)*n_line)
        print('dist1: {} r1: {}'.format(dist1,r1))
        print('dist2: {} r2: {}'.format(dist2,r2))
        if dist1 < r1 or dist2 < r2:
            return True
        else:
            return False
    elif not np.isclose(ny1,0):
        alpha = (nz1*ny2-nz2*ny1)/(nx2*ny1-nx1*ny2)
        beta  = (d1*ny2-d2*ny1)/(nx2*ny1-nx1*ny2)
        c     = nx1/ny1
        a_pt  = np.array([beta,-d1/ny1-beta*c,0])
        pt1   = np.array([x1,y1,z1])
        pt2   = np.array([x2,y2,z2])
        n_line= np.array([alpha,-(alpha*c+nz1/ny1),1])
        n_line=n_line/np.linalg.norm(n_line)
        dist1 = np.linalg.norm((pt1-a_pt)-np.dot(pt1-a_pt,n_line)*n_line)
        dist2 = np.linalg.norm((pt2-a_pt)-np.dot(pt2-a_pt,n_line)*n_line)
        print('dist1: {} r1: {}'.format(dist1,r1))
        print('dist2: {} r2: {}'.format(dist2,r2))
        if dist1 < r1 or dist2 < r2:
            return True
        else:
            return False
    else:
        return None
    """
    n = np.linspace(t1,t2)
    length = ((x2-x1)**2+(y2-y1)**2+(z2-z1)**2)**(1/2)
    if radius_buffer is not None:
        r1 += r1*radius_buffer
        r2 += r2*radius_buffer
    """
    contour_1 = vtk.vtkRegularPolygonSource()
    contour_2 = vtk.vtkRegularPolygonSource()
    contour_1.SetCenter([x1,y1,z1])
    contour_2.SetCenter([x2,y2,z2])
    contour_1.SetRadius(r1)
    contour_2.SetRadius(r2)
    contour_1.SetNormal([nx1,ny1,nz1])
    contour_2.SetNormal([nx2,ny2,nz2])
    contour_1.SetNumberOfSides(n_sides)
    contour_2.SetNumberOfSides(n_sides)
    contour_11 = vtk.vtkContourTriangulator()
    contour_22 = vtk.vtkContourTriangulator()
    contour_11.SetInputConnection(contour_1.GetOutputPort())
    contour_22.SetInputConnection(contour_2.GetOutputPort())
    collision = vtk.vtkCollisionDetectionFilter()
    collision.SetInputConnection(0,contour_11.GetOutputPort())
    collision.SetInputConnection(1,contour_22.GetOutputPort())
    transform = vtk.vtkTransform()
    matrix = vtk.vtkMatrix4x4()
    collision.SetTransform(0,transform)
    collision.SetMatrix(1,matrix)
    collision.SetCollisionModeToAllContacts()
    collision.GenerateScalarsOn()
    collision.Update()
    """
    contour_1 = pv.Disc(center=[x1,y1,z1],inner=0,outer=r1,normal=[nx1,ny1,nz1],r_res=5,c_res=20).triangulate()
    contour_2 = pv.Disc(center=[x2,y2,z2],inner=0,outer=r2,normal=[nx2,ny2,nz2],r_res=5,c_res=20).triangulate()
    collision,ncol = contour_1.collision(contour_2,contact_mode=0)
    #if collision.GetNumberOfContacts() > 0:
    if ncol > 0:
        """
        actors = []
        colors = vtk.vtkNamedColors()
        mapper_1 = vtk.vtkPolyDataMapper()
        mapper_2 = vtk.vtkPolyDataMapper()
        actor_1 = vtk.vtkActor()
        actor_2 = vtk.vtkActor()
        mapper_1.SetInputConnection(contour_11.GetOutputPort())
        mapper_2.SetInputConnection(contour_22.GetOutputPort())
        actor_1.SetMapper(mapper_1)
        actor_2.SetMapper(mapper_2)
        actor_1.GetProperty().SetColor(colors.GetColor3d('blue'))
        actor_2.GetProperty().SetColor(colors.GetColor3d('blue'))
        actors.append(actor_1)
        actors.append(actor_2)
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(colors.GetColor3d('white'))

        render_window = vtk.vtkRenderWindow()
        render_window.AddRenderer(renderer)
        render_window.SetWindowName('Pathline with Contours')

        interactor = vtk.vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)
        for actor in actors:
            renderer.AddActor(actor)
        render_window.Render()
        interactor.Start()
        """
        return True
    else:
        return False
    """
    #print('{},{},{}'.format(x1,y1,z1))
    p1 = Point3D(Float(x1),Float(y1),Float(z1))
    p2 = Point3D(Float(x2),Float(y2),Float(z2))
    a = Plane(p1,normal_vector=(Float(nx1),Float(ny1),Float(nz1)))
    b = Plane(p2,normal_vector=(Float(nx2),Float(ny2),Float(nz2)))
    intersect = a.intersection(b)[0]
    dist1 = intersect.distance(p1)
    dist2 = intersect.distance(p2)
    p_seg1 = intersect.perpendicular_line(p1)
    p_seg2 = intersect.perpendicular_line(p2)
    pi1 = intersect.intersection(p_seg1)[0]
    pi2 = intersect.intersection(p_seg2)[0]
    line_1 = np.array([[p1.x,p1.y,p1.z],[pi1.x,pi1.y,pi1.z]])
    line_2 = np.array([[p2.x,p2.y,p2.z],[pi2.x,pi2.y,pi2.z]])
    n1_vec = np.array([[p1.x,p1.y,p1.z],[p1.x+nx1,p1.y+ny1,p1.z+nz1]])
    n2_vec = np.array([[p2.x,p2.y,p2.z],[p2.x+nx2,p2.y+ny2,p2.z+nz2]])
    #print('dist1: {} r1: {}'.format(dist1,r1))
    #print('dist2: {} r2: {}'.format(dist2,r2))
    if dist1 < r1 or dist2 < r2:
        return True
    else:
        return False
    """
def contour_check(interp_xyz,interp_r,t0,t1,radius_buffer,n_sides=4):
    tl  = (t1 - t0)/10
    x1,y1,z1,r1      = splev(t0,interp_xyz[0])
    #_,r1            = splev(t0,interp_r[0])
    nx1,ny1,nz1,_ = splev(t0,interp_xyz[0],der=1)
    n1  = np.array([nx1,ny1,nz1])
    mag = np.linalg.norm(n1,axis=0)
    n1 = n1/mag
    x2,y2,z2,r2            = splev(t1,interp_xyz[0])
    #_,r2                  = splev(t1,interp_r[0])
    nx2,ny2,nz2,_ = splev(t1,interp_xyz[0],der=1)
    n2  = np.array([nx2,ny2,nz2])
    mag = np.linalg.norm(n2,axis=0)
    n2 = n2/mag
    collision = contour_collision(x2,y2,z2,nx2,ny2,nz2,r2,x1,y1,z1,nx1,ny1,nz1,r1,radius_buffer,interp_xyz,t0,t1,n_sides=n_sides)
    return collision

def contour_check_all(interp_xyz,interp_r,radius_buffer):
    ANGLE_THRESHOLD = 20
    sample_t = []
    #n = np.linspace(0,1,1000)
    desc1 = 'Extracting Loft Contours'
    desc1 = desc1+' '*(40-len(desc1))
    for kdx in tqdm(range(len(interp_xyz)),bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=2,leave=False,desc=desc1):
        tmp_t = []
        t_list = []
        t_list = np.linspace(0,1,10*len(interp_xyz[kdx][1]))
        #plot_path_contours(interp_xyz[kdx],interp_r[kdx],t_list)
        normal_list = []
        x,y,z,r = splev(0,interp_xyz[kdx][0])
        previous = [x,y,z,r]
        nx,ny,nz,_ = splev(0,interp_xyz[kdx][0],der=1)
        normal     = np.array([nx,ny,nz])
        normal     = normal/np.linalg.norm(normal)
        normal_list.append(normal)
        # Clean out all contours that provide little information
        tmp_t.append(t_list[0])
        desc2 = 'Cleaning Out Redundant Contours'
        desc2 = desc2+' '*(40-len(desc2))
        for ct in tqdm(t_list[1:-1],bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=3,leave=False,desc=desc2):
            x,y,z,r = splev(ct,interp_xyz[kdx][0])
            nx,ny,nz,_ = splev(ct,interp_xyz[kdx][0],der=1)
            normal     = np.array([nx,ny,nz])
            normal     = normal/np.linalg.norm(normal)
            angle      = abs((np.arccos(np.dot(normal_list[-1],normal.T))/np.pi)*180)
            dist       = ((previous[0]-x)**2+(previous[1]-y)**2+(previous[2]-z)**2)**(1/2)
            if angle > ANGLE_THRESHOLD or (dist>(r+previous[3])):
                normal_list.append(normal)
                tmp_t.append(ct)
                previous = [x,y,z,r]
        tmp_t.append(t_list[-1])
        #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_t)
        """
        if len(tmp_t) > 2:
            tmp_tmp_t = [tmp_t[0]]
            for ct in tqdm(tmp_t[1:-1],position=2,desc="Cleaning Consecutive Colliding Contours"):
                _,_,_,r = splev(ct,interp_xyz[kdx][0])
                if contour_check(interp_xyz[kdx],interp_r[kdx],tmp_tmp_t[-1],ct,r*0.25,n_sides=20):
                    if len(tmp_tmp_t) > 1:
                        tmp_tmp_t[-1] = (ct+tmp_tmp_t[-1])/2
                else:
                    tmp_tmp_t.append(ct)
            tmp_tmp_t.append(tmp_t[-1])
        else:
            tmp_tmp_t = tmp_t
        """
        tmp_tmp_t = tmp_t
        #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_t)
        if len(tmp_t) > 2:
            tmp_tmp_tmp_t = [tmp_tmp_t[0]]
            desc3 = 'Cleaning remaining collisions'
            desc3 = desc3+' '*(40-len(desc3))
            for ct in tqdm(tmp_tmp_t[1:-1],bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=3,leave=False,desc=desc3):
                _,_,_,r = splev(ct,interp_xyz[kdx][0])
                include = True
                for cct in tmp_tmp_t[1:-1]:
                    if cct==ct:
                        continue
                    if contour_check(interp_xyz[kdx],interp_r[kdx],ct,cct,r*0.25,n_sides=20):
                        include = False
                        break
                if include:
                    tmp_tmp_tmp_t.append(ct)
            tmp_tmp_tmp_t.append(tmp_t[-1])
        else:
            tmp_tmp_tmp_t = tmp_tmp_t
        #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_tmp_t)
        if len(tmp_tmp_t) > 2:
            repeat = True
            old_count = len(tmp_tmp_tmp_t)
            resample_count = 1
            while repeat:
                tmp_tmp_tmp_tmp_t = [tmp_tmp_tmp_t[0]]
                normal_list = []
                nx,ny,nz,_ = splev(0,interp_xyz[kdx][0],der=1)
                normal     = np.array([nx,ny,nz])
                normal     = normal/np.linalg.norm(normal)
                normal_list.append(normal)
                highlight = []
                count = 1
                swap_list = []
                total = len(tmp_tmp_tmp_t[1:-1])
                desc4 = 'Resampling {}'.format(resample_count)
                desc4 = desc4+' '*(40-len(desc4))
                with tqdm(total=total,bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=3,leave=False,desc=desc4) as pbar:
                    for ct_idx,ct in enumerate(tmp_tmp_tmp_t[1:-1]):
                        t_next = tmp_tmp_tmp_t[ct_idx + 2]
                        x,y,z,r = splev(ct,interp_xyz[kdx][0])
                        nx,ny,nz,_ = splev(ct,interp_xyz[kdx][0],der=1)
                        normal     = np.array([nx,ny,nz])
                        normal     = normal/np.linalg.norm(normal)
                        angle      = abs((np.arccos(np.dot(normal_list[-1],normal.T))/np.pi)*180)
                        if angle > ANGLE_THRESHOLD+5:
                            t_op,f_op = optimize_contour(tmp_tmp_tmp_tmp_t[-1],ct,interp_xyz[kdx],interp_r[kdx],0.25,n_sides=20)
                            tnx,tny,tnz,_ = splev(t_op,interp_xyz[kdx][0],der=1)
                            next_nx,next_ny,next_nz,_ = splev(t_next,interp_xyz[kdx][0],der=1)
                            t_op_normal     = np.array([tnx,tny,tnz])
                            t_op_normal     = t_op_normal/np.linalg.norm(t_op_normal)
                            angle0      = abs((np.arccos(np.dot(normal_list[-1],t_op_normal.T))/np.pi)*180)
                            t_next_normal     = np.array([next_nx,next_ny,next_nz])
                            t_next_normal     = t_next_normal/np.linalg.norm(t_next_normal)
                            angle1      = abs((np.arccos(np.dot(t_op_normal,t_next_normal.T))/np.pi)*180)
                            if f_op < 0 and (angle0 > ANGLE_THRESHOLD*0.5 and angle1 > ANGLE_THRESHOLD*0.5):
                                #t_op = (tmp_tmp_tmp_t[-1]+ct)/2
                                tmp_tmp_tmp_tmp_t.append(t_op)
                                if count+1 < len(tmp_tmp_tmp_t)-1:
                                    swap_list.append(count+1)
                                if count-1 > 0:
                                    swap_list.append(count-1)
                                highlight.append(count)
                                count += 1
                            tmp_tmp_tmp_tmp_t.append(ct)
                            count += 1
                        else:
                            tmp_tmp_tmp_tmp_t.append(ct)
                            count += 1
                        normal_list.append(normal)
                        pbar.update(1)
                tmp_tmp_tmp_tmp_t.append(tmp_tmp_t[-1])
                #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_tmp_tmp_t,highlight=highlight)
                if len(tmp_tmp_tmp_tmp_t) == old_count:
                    repeat = False
                #else:
                #    better_t,better_idx = swap_contour(interp_xyz[kdx],tmp_tmp_tmp_tmp_t,0.25)
                #    tmp_tmp_tmp_tmp_t[better_idx] = better_t
                #    plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_tmp_tmp_t,highlight=[better_idx])
                for sw in swap_list:
                    tmp_tmp_tmp_tmp_t[sw] = swap_contour(interp_xyz[kdx],tmp_tmp_tmp_tmp_t,sw,0.25)
                #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_tmp_tmp_t,highlight=swap_list)
                tmp_tmp_tmp_tmp_t,better_idx = swap_worst_contour(interp_xyz[kdx],tmp_tmp_tmp_tmp_t,0.25,thresh=2*ANGLE_THRESHOLD)
                #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_tmp_tmp_t,highlight=better_idx)
                old_count = len(tmp_tmp_tmp_tmp_t)
                tmp_tmp_tmp_t = tmp_tmp_tmp_tmp_t
                resample_count += 1
        else:
            tmp_tmp_tmp_tmp_t = tmp_tmp_tmp_t
        #plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_tmp_tmp_tmp_t,highlight=highlight)
        sample_t.append(tmp_tmp_tmp_tmp_t)
    return sample_t
"""
ORIGINAL CONTOUR CHECKING
def contour_check_all(interp_xyz,interp_r,radius_buffer):
    sample_t = []
    n = np.linspace(0,1,200)
    total_collisions = 0
    for kdx in tqdm(range(len(interp_xyz)),desc='Resolving Overlaps'):
        tmp_t  = []
        t_list = []
        t_list = np.linspace(0,1,4*len(interp_xyz[kdx][1]))
        collision_contours = set([])
        collision_free = set([])
        checking = set(list(range(1,len(t_list)-1)))
        #t_op,f_op = optimize_contour(t_range[first],t_range[second],contours)
        for idx in checking:
            collide = False
            for jdx in checking.difference(set([idx])):
                if contour_check(interp_xyz[kdx],interp_r[kdx],t_list[idx],t_list[jdx],radius_buffer,n_sides=20):
                    collision_contours.add(idx)
                    collision_contours.add(jdx)
                    collide = True
                    break
            if not collide:
                collision_free.add(idx)
            #checking = checking.difference(collision_contours)
            checking = checking.difference(collision_free)
        collision_free = list(collision_free)
        collision_free.sort()
        if len(collision_contours) > 0:
            pass
            plot_path_contours(interp_xyz[kdx],interp_r[kdx],t_list,highlight=list(collision_contours))
        tmp_t.append(t_list[0])
        for idx in range(len(collision_free)-1):

            if collision_free[idx+1] - collision_free[idx]>1:
                tmp_tmp_t = [t_list[collision_free[idx]],t_list[collision_free[idx+1]]]
                #nx,ny,nz,_,_,_,_ = splev(tmp_tmp_t,contours,der=1)
                #normals = np.array([nx,ny,nz]).T
                #tmp_tmp_angles = [np.dot(normals[i,:],normals[i+1,:]) for i in range(normals.shape[0]-1)]
                for jdx in range(collision_free[idx+1] - collision_free[idx]):
                    #max_diff = np.argmax(np.diff(tmp_tmp_angles))
                    nx,ny,nz,_ = splev(tmp_tmp_t,interp_xyz[kdx][0],der=1)
                    normals = np.array([nx,ny,nz]).T
                    tmp_tmp_angles = [np.dot(normals[i,:],normals[i+1,:]) for i in range(normals.shape[0]-1)]
                    max_diff = np.argmin(tmp_tmp_angles)
                    t_op,f_op = optimize_contour(tmp_tmp_t[max_diff],tmp_tmp_t[max_diff+1],interp_xyz[kdx],interp_r[kdx],radius_buffer,n_sides=20)
                    if f_op < 0:
                        tmp_tmp_t.append(t_op)
                        tmp_tmp_t.sort()
                    else:
                        break
                tmp_tmp_t.pop(-1)
                tmp_t.extend(tmp_tmp_t)
            else:
                tmp_t.append(t_list[collision_free[idx]])
        tmp_t.append(t_list[collision_free[-1]])
        tmp_t.append(t_list[-1])
        if len(collision_contours) > 0:
            pass
            plot_path_contours(interp_xyz[kdx],interp_r[kdx],tmp_t)
        if len(tmp_t) == 1:
            tmp_t.append(1)
        else:
            tmp_t[-1] = 1
        sample_t.append(tmp_t)
    print('Total Collisions Resolved: '+str(total_collisions))
    return sample_t
"""

def sv_data(interp_xyz,interp_r,radius_buffer=0.25):
    points  = []
    radii   = []
    normals = []
    sample_t = contour_check_all(interp_xyz,interp_r,radius_buffer)
    #fig = plt.figure()
    #ax = fig.add_subplot(111,projection='3d')
    for idx in range(len(interp_xyz)):
        x,y,z,r    = splev(np.array(sample_t[idx]),interp_xyz[idx][0])
        #_,r        = splev(np.array(sample_t[idx]),interp_r[idx][0])
        nx,ny,nz,_ = splev(np.array(sample_t[idx]),interp_xyz[idx][0],der=1)
        l  = np.linalg.norm(np.array([nx,ny,nz]),axis=0)
        nx = nx/l
        ny = ny/l
        nz = nz/l
        #ax.plot3D(x,y,z)
        #for i in range(len(x)):
        #    patch = Circle((0,0),r[i],facecolor='r')
        #    ax.add_patch(patch)
        #    pathpatch_2d_to_3d(patch,z=0,normal=np.array([nx[i],ny[i],nz[i]]))
        #    pathpatch_translate(patch,(x[i],y[i],z[i]))
        tmp_points = np.zeros((len(x),3))
        tmp_points[:,0] = x
        tmp_points[:,1] = y
        tmp_points[:,2] = z
        points.append(tmp_points.tolist())
        tmp_normals = np.zeros((len(nx),3))
        tmp_normals[:,0] = nx
        tmp_normals[:,1] = ny
        tmp_normals[:,2] = nz
        radii.append(r.tolist())
        normals.append(tmp_normals.tolist())
    #plt.show()
    return points,radii,normals

def plot_frames(frames):
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    for idx,f in enumerate(frames):
        f = np.array(f)
        x,y,z = (f[:,0],f[:,1],f[:,2])
        ax.plot3D(x,y,z)
    plt.show()

def plot_branches(data,branches):
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    for i,branch in enumerate(branches):
        for idx in branch:
            ax.plot3D([data[idx,0],data[idx,3]],
                      [data[idx,1],data[idx,4]],
                      [data[idx,2],data[idx,5]],label=str(i))
    plt.legend()
    plt.show()

def plot_path_contours(interp_xyz,interp_r,t_list,highlight=[]):
    disks = []
    lines = []
    points = vtk.vtkPoints()
    pts = []
    radii = []
    normals = []
    actors = []
    colors = vtk.vtkNamedColors()
    for t in t_list:
        x,y,z,r = splev(t,interp_xyz[0])
        #_,r     = splev(t,interp_r[0])
        nx,ny,nz,_ = splev(t,interp_xyz[0],der=1)
        points.InsertNextPoint([x,y,z])
        radii.append(r)
        normals.append([nx,ny,nz])
        pts.append([x,y,z])
    polyline = vtk.vtkPolyLine()
    polyline.GetPointIds().SetNumberOfIds(len(t_list))
    for i in range(len(t_list)):
        polyline.GetPointIds().SetId(i,i)
    cells = vtk.vtkCellArray()
    cells.InsertNextCell(polyline)
    polydata = vtk.vtkPolyData()
    polydata.SetPoints(points)
    polydata.SetLines(cells)
    mapper = vtk.vtkPolyDataMapper()
    actor = vtk.vtkActor()
    mapper.SetInputDataObject(polydata)
    actor.SetMapper(mapper)
    actor.GetProperty().SetColor(colors.GetColor3d('red'))
    actors.append(actor)
    for r in range(len(radii)):
        tmp_disk = vtk.vtkRegularPolygonSource()
        tmp_disk.SetRadius(radii[r])
        tmp_disk.SetCenter(pts[r])
        tmp_disk.SetNormal(normals[r])
        tmp_disk.SetNumberOfSides(100)
        mapper = vtk.vtkPolyDataMapper()
        actor = vtk.vtkActor()
        mapper.SetInputConnection(tmp_disk.GetOutputPort())
        actor.SetMapper(mapper)
        if r in highlight:
            actor.GetProperty().SetColor(colors.GetColor3d('blue'))
        else:
            actor.GetProperty().SetColor(colors.GetColor3d('red'))
        disks.append(tmp_disk)
        actors.append(actor)
    renderer = vtk.vtkRenderer()
    renderer.SetBackground(colors.GetColor3d('white'))

    render_window = vtk.vtkRenderWindow()
    render_window.AddRenderer(renderer)
    render_window.SetWindowName('Pathline with Contours')

    interactor = vtk.vtkRenderWindowInteractor()
    interactor.SetRenderWindow(render_window)
    for actor in actors:
        renderer.AddActor(actor)
    render_window.Render()
    interactor.Start()

def optimize_contour(t0,t1,interp_xyz,interp_r,radius_buffer,t_buffer=0.1,n_sides=50):
    nx0,ny0,nz0,_ = splev(t0,interp_xyz[0],der=1)
    nx1,ny1,nz1,_ = splev(t1,interp_xyz[0],der=1)
    n0 = np.array([nx0,ny0,nz0])
    n1 = np.array([nx1,ny1,nz1])
    n0 = n0/np.linalg.norm(n0) #.reshape(-1,1)
    n1 = n1/np.linalg.norm(n1) #.reshape(-1,1)
    t_bound_buff = (t1-t0)*t_buffer
    def func(t,interp_xyz=interp_xyz,interp_r=interp_r,n0=n0,n1=n1,t0=t0,t1=t1):
        x0,y0,z0,r0 = splev(t0,interp_xyz[0])
        x1,y1,z1,r1 = splev(t1,interp_xyz[0])
        x,y,z,r = splev(t,interp_xyz[0])
        nx,ny,nz,_ = splev(t,interp_xyz[0],der=1)
        n = np.array([nx,ny,nz])
        n = n/np.linalg.norm(n) #.reshape(-1,1)
        #print(-np.dot(n0,n)*np.dot(n1,n))
        n_avg = (n0+n1)/2
        a = np.dot(n_avg,n)
        first_angle = np.arccos(np.dot(n0,n))
        second_angle = np.arccos(np.dot(n1,n))
        #l0 = ((x0-x)**(2)+(y0-y)**(2)+(z0-z)**(2))**(1/2)
        #l1 = ((x1-x)**(2)+(y1-y)**(2)+(z1-z)**(2))**(1/2)
        mx  = (x0+x1)/2
        my  = (y0+y1)/2
        mz  = (z0+z1)/2
        lm1 = ((x0-x)**(2)+(y0-y)**(2)+(z0-z)**(2))**(1/2)
        lm2 = ((x1-x)**(2)+(y1-y)**(2)+(z1-z)**(2))**(1/2)
        #d   = (((x1-x0)**(2)+(y1-y0)**(2)+(z1-z0)**(2))**(1/2))/2
        #d   = d/2
        lm = abs(lm1-lm2)/(lm1+lm2)
        c = 0
        if contour_check(interp_xyz,interp_r,t0,t,radius_buffer=radius_buffer,n_sides=n_sides):
            c = 100
        if contour_check(interp_xyz,interp_r,t1,t,radius_buffer=radius_buffer,n_sides=n_sides):
            c += 100
        if np.sin(first_angle)*(r0+r) > lm1:
            c += 100
        if np.sin(second_angle)*(r1+r) > lm2:
            c += 100
        #return -(a*b+(1-abs(a-b))) + c # + abs(l0-l1)
        return -(a) + c + lm/100
    result = optimize.minimize_scalar(func,bounds=(t0+t_bound_buff,t1-t_bound_buff),method='bounded')
    return result.x,result.fun

def swap_worst_contour(interp_xyz,t_list,radius_buffer,thresh=20):
    normal_list = []
    nx,ny,nz,_ = splev(0,interp_xyz[0],der=1)
    normal     = np.array([nx,ny,nz])
    normal     = normal/np.linalg.norm(normal)
    normal_list.append(normal)
    #next_nx,next_ny,next_nz,_ = splev(t_list[1],interp_xyz[kdx][0],der=1)
    #next_normal     = np.array([next_nx,next_ny,next_nz])
    #next_normal     = next_normal/np.linalg.norm(next_normal)
    #angle = abs((np.arccos(np.dot(normal_list[-1],next_normal.T))/np.pi)*180)
    #scores = [angle]
    scores = []
    total = len(t_list[1:-1])
    desc5 = 'Computing swap scores'
    desc5 = desc5+' '*(40-len(desc5))
    with tqdm(total=total,bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=4,leave=False,desc=desc5) as pbar:
        for idx,ct in enumerate(t_list[1:-1]):
            next_t            = t_list[idx+2]
            nx,ny,nz,_        = splev(ct,interp_xyz[0],der=1)
            normal            = np.array([nx,ny,nz])
            normal            = normal/np.linalg.norm(normal)
            angle0            = abs((np.arccos(np.dot(normal_list[-1],normal.T))/np.pi)*180)
            next_nx,next_ny,next_nz,_ = splev(next_t,interp_xyz[0],der=1)
            t_next_normal     = np.array([next_nx,next_ny,next_nz])
            t_next_normal     = t_next_normal/np.linalg.norm(t_next_normal)
            angle1            = abs((np.arccos(np.dot(normal,t_next_normal.T))/np.pi)*180)
            scores.append(angle0+angle1)
            normal_list.append(normal)
            pbar.update(1)
    worst_idx = np.argwhere(np.array(scores)>thresh).flatten()+1
    worst_idx = worst_idx.tolist()
    for wi in worst_idx:
        t   = t_list[wi]
        t0,f0 = optimize_contour(t_list[wi-1],t_list[wi+1],interp_xyz,None,radius_buffer,t_buffer=0.1,n_sides=50)
        if f0 < 0:
            t_list[wi] = t0
    return t_list,worst_idx

def swap_contour(interp_xyz,t_list,c_idx,radius_buffer):
    value = t_list[c_idx]
    t0,f0 = optimize_contour(t_list[c_idx-1],t_list[c_idx+1],interp_xyz,None,radius_buffer,t_buffer=0.1,n_sides=50)
    if f0 < 0:
        value = t0
    return value

def truncate(data,radius=None,indicies=None):
    if radius is None and indicies is None:
        radius = np.median(data[:,21])
    branches = get_branches(data)
    include = []
    exclude = []
    for branch in branches:
        tmp_include = []
        tmp_exclude = []
        for idx,vdx in enumerate(branch):
            if indicies is None:
                if data[vdx,21] >= radius:
                    tmp_include.append(vdx)
                else:
                    tmp_exclude.append(vdx)
            else:
                if vdx in indicies:
                    tmp_include.append(vdx)
                else:
                    tmp_exclude.append(vdx)
        if len(tmp_include) > 1:
            include.append(tmp_include)
        if len(tmp_exclude) > 0:
            exclude.append(tmp_exclude)
    return include, exclude

def get_truncated_interpolated_sv_data(data,radius=None,indicies=None):
    branches,_ = truncate(data,radius=radius,indicies=indicies)
    points   = get_points(data,branches)
    #print("Points: {}".format(points[0]))
    radii    = get_radii(data,branches)
    #print("Radii:  {}".format(radii[0]))
    normals  = get_normals(data,branches)
    path_frames = []
    for idx in range(len(branches)):
        frames = []
        for jdx in range(len(points[idx])):
            frame = []
            frame.extend(points[idx][jdx])
            frame.append(radii[idx][jdx])
            frame.extend(normals[idx][jdx])
            frames.append(frame)
        path_frames.append(frames)
    interp_xyz = []
    interp_r   = []
    interp_n   = []
    interp_xyzr = []
    for idx in range(len(branches)):
        p = np.array(points[idx]).T
        r = np.array(radii[idx]).T
        n = np.array(normals[idx]).T
        if len(points[idx]) == 2:
            interp_xyz.append(splprep(p,k=1,s=0))
            rr = np.vstack((interp_xyz[-1][1],r))
            interp_r.append(splprep(rr,k=1,s=0))
            xyzr = np.vstack((p,r))
            interp_xyzr.append(splprep(xyzr,k=1,s=0))
            #interp_n.append(splprep(n,k=1,s=0))
        elif len(points[idx]) == 3:
            interp_xyz.append(splprep(p,k=2,s=0))
            rr = np.vstack((interp_xyz[-1][1],r))
            interp_r.append(splprep(rr,k=2,s=0))
            xyzr = np.vstack((p,r))
            interp_xyzr.append(splprep(xyzr,k=2,s=0))
            #interp_n.append(splprep(n,k=2,s=0))
        else:
            interp_xyz.append(splprep(p,s=0))
            rr = np.vstack((interp_xyz[-1][1],r))
            interp_r.append(splprep(rr,s=0))
            xyzr = np.vstack((p,r))
            interp_xyzr.append(splprep(xyzr,s=0))
            #interp_n.append(splprep(n,s=0))

    return interp_xyz,interp_r,interp_n,path_frames,branches,interp_xyzr



# Code for result figure 1
#
# What are we trying to show?
# --Build time for models
# --The models themselves (connected but not lofted)
# --Diameters versus depth

import svcco
from copy import deepcopy
import numpy as np
import os
import pickle
import time
import pyvista as pv
cube = pv.Cube().triangulate().subdivide(3)

A = [1]#,2,3,4,5]
B = [1]#,1,1,1,1]
S = []
C = []
for i in range(len(A)):
    if i == 4:
        sph = svcco.bumpy_sphere(samples=35,scale=3,a=A[i],b=B[i])
    else:
        sph = svcco.bumpy_sphere(samples=20,scale=3,a=A[i],b=B[i])
    s = svcco.surface()
    #s.set_data(np.array(10*cube.points),normals=cube.point_normals)
    s.set_data(sph)
    s.solve(regularize=True)
    #s.solve()
    s.build()
    #svcco.plot_volume(s)
    #print(s.DD[0]((s.x_range[1]+0.1,s.y_range[1]+0.1,s.z_range[1]+0.1,len(s.patches))))
    if s.DD[0]((s.x_range[1]+0.1,s.y_range[1]+0.1,s.z_range[1]+0.1,len(s.patches))) < 0:
        s.normals = -s.normals
        s.solve(regularize=True)
        #s.solve()
        s.build()
        #svcco.plot_volume(s)
    if i == 0:
        parent_convexity = s.surface_area/s.volume
        convexity = 1
    else:
        convexity = (s.surface_area/s.volume)/parent_convexity
    S.append(s)
    C.append(convexity)

s = svcco.surface()
s.set_data(np.array(10*cube.points),normals=cube.point_normals)
s.solve()
s.build()
if s.DD[0]((s.x_range[1]+0.1,s.y_range[1]+0.1,s.z_range[1]+0.1,len(s.patches))) < 0:
    s.normals = -s.normals
    s.solve()
    s.build()
S.append(s)
C.append(1)

V = [100,900,1000]#,2000]
repeats = 2
DATA = []
T = []
t = svcco.tree()
t.set_boundary(S[0])
t.set_root()
t.n_add(10)
os.mkdir(os.getcwd()+os.sep+'figure1')
os.chdir(os.getcwd()+os.sep+'figure1')
for n,surf in enumerate(S):
    s_data = []
    t_data = []
    for j in range(repeats):
        tree_data = []
        time_data = []
        t = svcco.tree()
        t.set_boundary(surf)
        t.set_root()
        for i in range(len(V)):
            start = time.time()
            t.n_add(V[i])
            end = time.time() - start
            time_data.append(end)
            tree_data.append(deepcopy(t.data))
            t.show(surface=True,save=True,name='{}_{}_{}'.format('S'+str(n)+str(1),str(j),str(V[i])))
        s_data.append(tree_data)
        t_data.append(time_data)
    DATA.append(s_data)
    T.append(t_data)
w = open(os.getcwd()+os.sep+'figure_1_data.pkl','wb+')
pickle.dump(DATA,w)
w.close()
wc = open(os.getcwd()+os.sep+'convexity.pkl','wb+')
pickle.dump(C,wc)
wc.close()
wt = open(os.getcwd()+os.sep+'time.pkl','wb+')
pickle.dump(T,wt)



import numpy as np
from scipy import spatial
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
from matplotlib import cm
from tqdm import tqdm

def generate_bases(points,quiet=True):
    n_points = points.shape[0]
    dim = points.shape[1]
    int_dim = dim - 1
    tangent_basis = np.zeros((dim,int_dim,n_points))
    normal_basis = np.zeros((dim,dim-int_dim,n_points))
    K = int(max(20, dim*np.log10(n_points)))
    KDT = spatial.KDTree(points)
    dist, idx = KDT.query(points,k=K+1)
    if quiet:
        for i in range(n_points):
            tmp = points[idx[i,:],:]
            tmp = tmp - np.ones((K+1,1))@tmp[0,:].reshape(1,-1)
            tmp = tmp[1:,:]
            U, S, H = np.linalg.svd(tmp)
            H = H.T
            diag_idx = S.argsort()[::-1]
            sort_diag = np.sort(S)[::-1]
            H = H[:,diag_idx]
            tangent_basis[:,:,i] = H[:,:int_dim]
            normal_basis[:,:,i] = H[:,dim-1].reshape(-1,1)
    else:
        for i in tqdm(range(n_points),desc='Generating bases       '):
    	    tmp = points[idx[i,:],:]
    	    tmp = tmp - np.ones((K+1,1))@tmp[0,:].reshape(1,-1)
    	    tmp = tmp[1:,:]
    	    U, S, H = np.linalg.svd(tmp)
    	    H = H.T
    	    diag_idx = S.argsort()[::-1]
    	    sort_diag = np.sort(S)[::-1]
    	    H = H[:,diag_idx]
    	    tangent_basis[:,:,i] = H[:,:int_dim]
    	    normal_basis[:,:,i] = H[:,dim-1].reshape(-1,1)
    return tangent_basis,normal_basis,idx,KDT

def estimate_weingarten_map(points,quiet=True):
    n_points = points.shape[0]
    dim = points.shape[1]
    tangent_basis, normal_basis, idx, KDT = generate_bases(points,quiet=quiet)
    tangent_dim = tangent_basis.shape[1]
    normal_dim = normal_basis.shape[1]
    weingarten_map = np.zeros((tangent_dim,tangent_dim,
                               normal_dim,n_points))
    if quiet:
        for i in range(n_points):
            for j in range(normal_dim):
                tmp_point = points[i,:]
                tmp_tangent_basis = tangent_basis[:,:,i]
                tmp_normal_basis = normal_basis[:,j,i]
                tmp_neighborhood = idx[i,:]
                tmp_neighborhood_size = len(tmp_neighborhood)
                tmp_local_normals = normal_basis[:,:,tmp_neighborhood]
                tmp_normal_extension = np.zeros((tmp_neighborhood_size,dim))
                for k in range(tmp_neighborhood_size):
                    projection = tmp_local_normals[:,:,k]@\
                                 tmp_local_normals[:,:,k].T@\
                                 tmp_normal_basis.reshape(-1,1)
                    tmp_normal_extension[k,:] = projection.T
                tmp_diff_normal = np.zeros((tmp_neighborhood_size-1,dim))
                tmp_diff_position = np.zeros((tmp_neighborhood_size-1,dim))
                for k in range(tmp_neighborhood_size-1):
                    tmp_diff_normal[k,:] = tmp_normal_extension[k+1,:] -\
                                           tmp_normal_basis.reshape(1,-1)
                    tmp_diff_position[k,:] = points[tmp_neighborhood[k+1],:] -\
                                             tmp_point
                tmp_normal_projection = tmp_diff_normal@tmp_tangent_basis
                tmp_position_projection = tmp_diff_position@tmp_tangent_basis
                A = -(np.linalg.inv(tmp_position_projection.T@\
                                    tmp_position_projection))@\
                                    tmp_position_projection.T@\
                                    tmp_normal_projection
                weingarten_map[:,:,j,i] = (1/2) * (A+A.T)
    else:
        for i in tqdm(range(n_points),desc='Building Weingarten Map'):
            for j in range(normal_dim):
                tmp_point = points[i,:]
                tmp_tangent_basis = tangent_basis[:,:,i]
                tmp_normal_basis = normal_basis[:,j,i]
                tmp_neighborhood = idx[i,:]
                tmp_neighborhood_size = len(tmp_neighborhood)
                tmp_local_normals = normal_basis[:,:,tmp_neighborhood]
                tmp_normal_extension = np.zeros((tmp_neighborhood_size,dim))
                for k in range(tmp_neighborhood_size):
                    projection = tmp_local_normals[:,:,k]@\
                                 tmp_local_normals[:,:,k].T@\
                                 tmp_normal_basis.reshape(-1,1)
                    tmp_normal_extension[k,:] = projection.T
                tmp_diff_normal = np.zeros((tmp_neighborhood_size-1,dim))
                tmp_diff_position = np.zeros((tmp_neighborhood_size-1,dim))
                for k in range(tmp_neighborhood_size-1):
                    tmp_diff_normal[k,:] = tmp_normal_extension[k+1,:] -\
                                           tmp_normal_basis.reshape(1,-1)
                    tmp_diff_position[k,:] = points[tmp_neighborhood[k+1],:] -\
                                              tmp_point
                tmp_normal_projection = tmp_diff_normal@tmp_tangent_basis
                tmp_position_projection = tmp_diff_position@tmp_tangent_basis
                A = -(np.linalg.inv(tmp_position_projection.T@\
                                    tmp_position_projection))@\
                                    tmp_position_projection.T@\
                                    tmp_normal_projection
                weingarten_map[:,:,j,i] = (1/2) * (A+A.T)
    return weingarten_map, idx, KDT

def estimate_gaussian_curvature(points,quiet=True):
    n_points = points.shape[0]
    guassian_curvature = np.zeros((n_points,1))
    w_map, idx, KDT = estimate_weingarten_map(points,quiet=quiet)
    for i in range(n_points):
        guassian_curvature[i] = np.linalg.det(w_map[:,:,0,i])
    return guassian_curvature, idx, KDT

def show_gaussian_curvature(points):
    guassian_curvature, idx, KDT = estimate_gaussian_curvature(points)
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    coolwarm = cm.get_cmap('coolwarm',1000)
    gc_min = min(guassian_curvature)
    gc_max = max(guassian_curvature)
    colors = coolwarm(guassian_curvature)
    cax = ax.scatter3D(points[:,0],points[:,1],points[:,2],cmap=coolwarm,c=guassian_curvature)
    cbar = fig.colorbar(cax,ax=ax,ticks=[gc_min,gc_max])
    return fig,ax
"""
def bumpy_sphere(samples=10,scale=5,a=3,b=1):
    data = np.zeros((samples**2,3))
    theta = np.linspace(0,2*np.pi,num=samples)
    phi = np.linspace(0,np.pi,num=samples)
    count = 0
    for t in theta:
        for p in phi:          
            r = scale + np.cos(a*t)*np.sin(b*p)
            data[count,0] = r*np.cos(t)*np.sin(p)
            data[count,1] = r*np.sin(t)*np.sin(p)
            data[count,2] = r*np.cos(p)
            count += 1
    final_data = []
    for i in range(data.shape[0]):
        if i == 0:
            final_data.append(data[i,:])
            continue
        add = True
        for j in range(i):
            if np.all(np.isclose(data[i,:],data[j,:])):
                add = False
        if add:
            final_data.append(data[i,:])
    return np.array(final_data)
"""



# rotate connections to avoid gross collisions
import numpy as np
from copy import deepcopy
from ..collision.collision import *

def rotate(R0,R1,P0,P1,theta):
    theta = (theta/180)*np.pi
    Rvector = R1-R0
    Rvector = Rvector/np.linalg.norm(Rvector)
    vector = P1-P0
    vector = vector/np.linalg.norm(vector)
    vector_parallel = (np.dot(vector,Rvector.T)/np.dot(Rvector,Rvector.T))*Rvector
    vector_perp     = vector-vector_parallel
    pn = np.linalg.norm(vector_perp)
    wn = np.linalg.norm(w)
    if pn < 0.1 or wn < 0.1:
        print(pn)
        print(wn)
        print(vector_perp)
        print(w)
    w      = np.cross(Rvector,vector_perp)
    x1     = np.cos(theta)/np.linalg.norm(vector_perp)
    x2     = np.sin(theta)/np.linalg.norm(w)
    rotated_vector = np.linalg.norm(vector_perp)*(x1*vector_perp+x2*w)+vector_parallel
    new_point = P0+rotated_vector*(np.linalg.norm(P1-P0))
    return new_point

def angle(P0,P1,P2):
    vector_1 = P0 - P1
    vector_2 = P2 - P1
    vector_1 = vector_1/np.linalg.norm(vector_1)
    vector_2 = vector_2/np.linalg.norm(vector_2)
    result   = np.arccos(np.dot(vector_1,vector_2)/np.pi)*180
    return result

def rotate_terminals(forest_copy):
    ANGLE_STEP = 5
    for ndx in range(forest_copy.number_of_networks):
        for tdx in range(forest_copy.trees_per_network):
            A = forest_copy.assignments[ndx][tdx]
            T = forest_copy.networks[ndx][tdx].data
            C = forest_copy.connections[ndx][0]
            for edx in range(len(forest_copy.connections[ndx][0])):
                if T[edx,17] < 0:
                    #Terminal root (cannot rotate)
                    continue
                ANGLE = angle(T[A[edx],0:3],T[A[edx],3:6],C[edx])
                attempt = 0
                parent  = int(T[A[edx],17])
                score = [angle]
                pts   = [deepcopy(T[A[edx],3:6])]
                while ANGLE < 90 and attempt < 10:
                    R0 = T[parent,0:3]
                    R1 = T[parent,3:6]
                    P0 = T[A[edx],0:3]
                    P1 = T[A[edx],3:6]
                    T[A[edx],3:6] = rotate(R0,R1,P0,P1,ANGLE_STEP)
                    ANGLE = angle(T[A[edx],0:3],T[A[edx],3:6],C[edx])
                    score.append(ANGLE)
                    pts.append(deepcopy(T[A[edx],3:6]))
                    attempt += 1
                if attempt == 10:
                    print("Fail")
                best = np.argmax(score)
                T[A[edx],3:6] = pts[best]
                T[A[edx],12:15] = T[A[edx],3:6]-T[A[edx],0:3]/np.linalg.norm(T[A[edx],3:6]-T[A[edx],0:3])



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def bumpy_sphere(samples=10,scale=5,a=3,b=1):
    data = np.zeros((samples**2,3))
    theta = np.linspace(0,2*np.pi,num=samples)
    phi = np.linspace(0,np.pi,num=samples)
    #P,T = np.meshgrid(theta,phi)
    count = 0
    for t in theta:
        for p in phi:
            r = scale + np.sin(a*t)*np.sin(b*p)
            data[count,0] = r*np.cos(t)*np.sin(p)
            data[count,1] = r*np.sin(t)*np.sin(p)
            data[count,2] = r*np.cos(p)
            count += 1
    final_data = []
    #def bs(t,p,a=a,b=b,scale=scale):
    #    r = scale + np.sin(a*t)*np.sin(b*p)
    #    x = r*np.cos(t)*np.sin(p)
    #    y = r*np.sin(t)*np.sin(p)
    #    z = r*np.cos(p)
    #    return x,y,z
    #x,y,z = bs(T,P)
    #x = x.flatten()
    #y = y.flatten()
    #z = z.flatten()
    #fig = plt.figure()
    #ax = fig.add_subplot(111,projection='3d')
    #ax.scatter3D(x,y,z)
    #plt.show()
    #final_data = np.array([x,y,z]).T
    #print(final_data.shape)
    #final_data = np.unique(final_data,axis=0)
    for i in range(data.shape[0]):
        if i == 0:
            final_data.append(data[i,:])
            continue
        add = True
        for j in range(i):
            if np.all(np.isclose(data[i,:],data[j,:])):
                add = False
        if add:
            final_data.append(data[i,:])
    return np.array(final_data)



import numpy as np
import numba as nb

@nb.jit(nopython=True)
def H(K00,K01,K11,lam):
    if lam > 0:
        inv = np.linalg.inv(np.eye(K00.shape[0])+lam*K00)
        H_matrix = K11 - (lam*K01.T)@inv@K01
    else:
        H_matrix = K11
    return H_matrix



import numpy as np
import pyvista as pv

def rotate(R0,R1,P0,P1,theta):
    theta = (theta/180)*np.pi
    Rvector = R1-R0
    Rvector = Rvector/np.linalg.norm(Rvector)
    vector = P1-P0
    vector = vector/np.linalg.norm(vector)
    vector_parallel = (np.dot(vector,Rvector.T)/np.dot(Rvector,Rvector.T))*Rvector
    vector_perp     = vector-vector_parallel
    w      = np.cross(Rvector,vector_perp)
    x1     = np.cos(theta)/np.linalg.norm(vector_perp)
    x2     = np.sin(theta)/np.linalg.norm(w)
    rotated_vector = np.linalg.norm(vector_perp)*(x1*vector_perp+x2*w)+vector_parallel
    new_point = P0+rotated_vector*(np.linalg.norm(P1-P0))
    return new_point

def line(P0,P1):
    mesh = pv.Line(P0,P1)
    return mesh

def disk(R0,R1,P0,P1):
    norm = R1-R0
    norm = norm/np.linalg.norm(norm)
    L = np.linalg.norm(P1-P0)
    mesh = pv.Disc(center=R1,normal=norm,outer=L,inner=0)
    return mesh

def test():
    R0 = np.random.random(3).reshape(1,-1)
    R1 = np.random.random(3).reshape(1,-1)
    P0 = R1
    P1 = np.random.random(3).reshape(1,-1)
    theta = 180*np.random.random(1)
    plotter = pv.Plotter()
    axis = line(R0,R1)
    original = line(P0,P1)
    P_new = rotate(R0,R1,P0,P1,theta)
    new = line(P0,P_new)
    d = disk(R0,R1,P0,P1)
    plotter.add_mesh(d,opacity=0.4)
    plotter.add_mesh(axis,color='black')
    plotter.add_mesh(original,color='r')
    plotter.add_mesh(new,color='g')
    plotter.show()
    return



import numpy as np
import numba as nb
from functools import partial

@nb.jit(nopython=True,cache=True)
def fast_point(a,b,n=np.array([1.0,1.0,1.0],dtype=np.float64),
          e1=np.array([0.0,0.0,0.0],dtype=np.float64),
          p_0=np.array([0.0,0.0,0.0],dtype=np.float64)):
    if n[2] == -1:
        e1 = np.array([-1,0,0],dtype=np.float64)
        e2 = np.array([0,-1,0],dtype=np.float64)
    else:
        e2 = np.cross(e1,n)
    p = np.add(p_0,a.reshape((-1,1))*e1)
    p = np.add(p,b.reshape((-1,1))*e2)
    return p

a = np.array([0.5,0.6],dtype=np.float64)
b = np.array([0.7,0.8],dtype=np.float64)

#@nb.jit(nopython=True,cache=True)
def fast_local_function(data,terminal_point,edge,
                   gamma,nu,Qterm,Pperm,Pterm,precomputed=None):
    # repository for calculated and adjusted values
    reduced_resistances = []
    reduced_lengths = []
    bifurcations = []
    flows = [] # should only be changing path (not alt paths)
    main_idx = [] # index 0: data[:,15]; index 1: data[:,16]
    alt_idx = []
    new_scale = [] # compound bifurcations
    alt_scale = []
    # begin crawling through tree
    total_flow = data[edge,22]*2+Qterm*2
    triad = ((data[edge,22]+Qterm)*data[edge,0:3]+data[edge,22]*data[edge,3:6]+Qterm*terminal_point)/total_flow
    uu = data[edge,0:3] - terminal_point
    uu = uu/np.linalg.norm(uu)
    vv = data[edge,3:6] - terminal_point
    vv = vv/np.linalg.norm(vv)
    n = np.cross(vv,uu)
    if precomputed is None:
        #local_points = fast_point(a,b,n=n,e1=uu,p_0=triad)
        return
    else:
        local_points = precomputed
    #local_points_v2 = fast_point(a,b,n=n,e1=uu,p_0=triad)
    #print(local_points_v2)
    #print(local_points_v2.shape)
    #print(precomputed)
    #print(precomputed.shape)
    upstream = data[edge,0:3]
    downstream = data[edge,3:6]
    upstream_length = np.sqrt(np.sum(np.square(local_points-upstream),axis=1))
    downstream_length = np.sqrt(np.sum(np.square(local_points-downstream),axis=1))
    terminal_length = np.sqrt(np.sum(np.square(local_points-terminal_point),axis=1))
    # Begin calculating reduced resistance and bifurcation ratios
    R_terminal = (8*nu/np.pi)*terminal_length
    R_terminal_sister = (8*nu/np.pi)*downstream_length + (data[edge,25]-(8*nu/np.pi)*data[edge,20])
    f_terminal = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(gamma/4)) ** (-1/gamma)
    f_terminal_sister = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(-gamma/4)) ** (-1/gamma)
    R_0 = (8*nu/np.pi)*upstream_length + ((f_terminal**4)/R_terminal +
                                          (f_terminal_sister**4)/R_terminal_sister)**-1
    reduced_resistances.append(R_0)
    L_0 = f_terminal ** 2 * terminal_length + f_terminal_sister ** 2 * (downstream_length + data[edge,27])
    reduced_lengths.append(L_0)
    flows.append(data[edge,22]+Qterm)
    main_idx.append(edge)
    bifurcations.append(f_terminal_sister) # 0 -> 15; 1-> 16
    bifurcations.append(f_terminal)
    if edge == 0:
        # on the terminating round the new terminal is always 16 and the existing is 15
        #bifurcations.append([f_terminal_sister,f_terminal])
        L_0 = upstream_length + L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        alt_idx.append(-1)
        new_scale.append(np.ones(f_terminal.shape))
        alt_scale.append(np.ones(f_terminal.shape))
        return (R_0,L_0,f_terminal,f_terminal_sister,None,local_points,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17])
    if int(data[edge,15]) == previous:
        alt = int(data[edge,16])
        alt_idx.append(alt)
        position = 0
        #bifurcations.append([f_terminal,f_terminal_sister])
    else:
        alt = int(data[edge,15])
        alt_idx.append(alt)
        position = 1
        #bifurcations.append([f_terminal_sister,f_terminal])
    f_changed = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (gamma/4)) ** (-1/gamma)
    f_stagnant = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (-gamma/4)) ** (-1/gamma)
    f_terminal = f_terminal*f_changed
    f_terminal_sister = f_terminal_sister*f_changed
    f_parent = f_changed
    R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4) /R_0 + (f_stagnant ** 4) /data[alt,25]) ** -1
    L_0 = f_changed ** 2 *(upstream_length+L_0) + f_stagnant ** 2 * (data[alt,20] + data[alt,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22]+Qterm)
    main_idx.append(edge)
    new_scale.append(f_changed)
    alt_scale.append(f_stagnant)
    if position == 0:
        bifurcations.append(f_changed)
        bifurcations.append(f_stagnant)
    else:
        bifurcations.append(f_stagnant)
        bifurcations.append(f_changed)
    if edge == 0:
        L_0 = data[edge,20]+L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,local_points,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17])
    while edge >= 0:
        if int(data[edge, 15]) == previous:
            alt = int(data[edge,16])
            alt_idx.append(alt)
            position = 0
        else:
            alt = int(data[edge,15])
            alt_idx.append(alt)
            position = 1
        f_changed = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (gamma/4)) ** (-1/gamma)
        f_stagnant = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (-gamma/4)) ** (-1/gamma)
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_parent*f_changed
        R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4 / R_0) + (f_stagnant ** 4 / data[alt, 25])) ** -1
        L_0 = f_changed ** 2 * (data[previous, 20] + L_0) + f_stagnant ** 2 *(data[alt, 20] + data[alt, 27])
        reduced_resistances.append(R_0)
        reduced_lengths.append(L_0)
        flows.append(data[edge,22]+Qterm)
        main_idx.append(edge)
        if edge >= 0:
            for j in range(len(new_scale)):
                new_scale[j] = new_scale[j]*f_changed
                alt_scale[j] = alt_scale[j]*f_changed
            new_scale.append(f_changed)
            #alt_scale = [bif*f_changed for bif in alt_scale]
            alt_scale.append(f_stagnant)
        #else:
        #    new_scale.append(np.ones(f_changed.shape))
        #    alt_scale.append(np.ones(f_changed.shape))
        if position == 0:
            bifurcations.append(f_changed)
            bifurcations.append(f_stagnant)
        else:
            bifurcations.append(f_stagnant)
            bifurcations.append(f_changed)
        previous = edge
        edge = int(data[edge,17])
    L_0 = data[0,20] + L_0
    R_0 = (((data[0, 22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
    return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,local_points,reduced_resistances,
            reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
            R_terminal_sister)



solid="""###############################
# INITIALIZE MODELING KERNEL
###############################
def robust_union(model_1,model_2):
    modeler = modeling.Modeler(modeling.Kernel.POLYDATA)
    model_1_be = bad_edges(model_1)
    model_2_be = bad_edges(model_2)
    print("Model 1 Bad Edges: {{}}\\n Model 2 Bad Edges: {{}}".format(model_1_be,model_2_be))
    if model_1_be == 0 and model_2_be == 0:
        unioned_model = modeler.union(model_1,model_2)
        unioned_model = clean(unioned_model)
        unioned_model = norm(unioned_model)
        if bad_edges(unioned_model) > 0:
            print('Unioned Model Bad Edges: {{}}'.format(bad_edges(unioned_model)))
            print('Filling')
            unioned_model = fill(unioned_model)
            print('Unioned Model Bad Edges: {{}}'.format(bad_edges(unioned_model)))
            print('Cleaning')
            unioned_model = clean(unioned_model)
            print('Unioned Model Bad Edges: {{}}'.format(bad_edges(unioned_model)))
            unioned_model = tri(unioned_model)
            print('Unioned Model Bad Edges: {{}}'.format(bad_edges(unioned_model)))
        print('union successful')
        return unioned_model
    else:
        print('1 or both models have bad edges.')
        unioned_model = modeler.union(model_1,model_2)
        unioned_model = clean(unioned_model)
        unioned_model = norm(unioned_model)
        return unioned_model

def union_all(solids,n_cells=100):
    for i in range(len(solids)):
        solids[i] = norm(solids[i])
        solids[i] = remesh(solids[i])
        solids[i] = remesh_caps(solids[i])
    joined = robust_union(solids[0],solids[1])
    for i in range(2,len(solids)):
        print("UNION NUMBER: "+str(i)+"/"+str(len(solids)-1))
        joined = robust_union(joined,solids[i])
        if joined is None:
            print("unioning failed")
            return None,True
    print("unioning passed")
    return joined,False

unioned_model,terminating = union_all(capped_vessels)
model = modeling.PolyData()
tmp = unioned_model.get_polydata()
NUM_CAPS = {}
############################
# COMBINE FACES
############################
if not terminating:
    model.set_surface(tmp)
    model.compute_boundary_faces({})
    caps = model.identify_caps()
    ids = model.get_face_ids()
    walls = [ids[i] for i,x in enumerate(caps) if not x]
    while len(walls) > 1:
        target = walls[0]
        lose = walls[1]
        model.combine_faces(target,[lose])
        #combined = mesh_utils.combine_faces(model.get_polydata(),target,lose)
        #model.set_surface(combined)
        ids = model.get_face_ids()
        caps = model.identify_caps()
        walls = [ids[i] for i,x in enumerate(caps) if not x]
        print(walls)
    ids = model.get_face_ids()
    if {}:
        dmg.add_model('{}',model)
    if len(ids) > NUM_CAPS:
        face_cells = []
        for idx in ids:
            face = model.get_face_polydata(idx)
            cells = face.GetNumberOfCells()
            print(cells)
            face_cells.append(cells)
        data_to_remove = len(ids) - NUM_CAPS
        remove_list = []
        for i in range(data_to_remove):
            remove_list.append(ids[face_cells.index(min(face_cells))])
            face_cells[face_cells.index(min(face_cells))] += 1000
        print(remove_list)
        while len(remove_list) > 0:
            target = walls[0]
            lose = remove_list.pop(-1)
            model.combine_faces(target,[lose])
            #combined = mesh_utils.combine_faces(model.get_polydata(),target,lose)
            #model.set_surface(combined)
            print(remove_list)
        print(model.get_face_ids())
    ###############################
    # LOCAL SMOOTHING (not included)
    ###############################
    #smoothing_params = {{'method':'constrained', 'num_iterations':5, 'constrain_factor':0.2, 'num_cg_solves':30}}
    smooth_model = model.get_polydata()
    for idx, contour_set in enumerate(contour_list):
         if idx == 0:
              continue
         smoothing_params = {{'method':'constrained', 'num_iterations':3, 'constrain_factor':0.1+(0.9*(1-contour_set[0].get_radius()/contour_list[0][0].get_radius())), 'num_cg_solves':30}}
         smooth_model = geometry.local_sphere_smooth(smooth_model,contour_set[0].get_radius()*2,contour_set[0].get_center(),smoothing_params)
         print('local sphere smoothing {{}}'.format(idx))
    model.set_surface(smooth_model)

model = clean(model)
#model = remesh_wall(model,walls[0])
"""


old_solid="""###############################
# INITIALIZE MODELING KERNEL
###############################
modeler = modeling.Modeler(modeling.Kernel.POLYDATA)
walls = []
faces = []
refined = []
better = []
n_cells = {}
###############################
# REMESH FACES
###############################
cap_model = modeling.PolyData()
for vessel in capped_solids:
    faces = vessel.identify_caps()
    face_ids = [i+1 for i,x in enumerate(faces) if x]
    tmp = mesh_utils.remesh(vessel.get_polydata(),hmin={},hmax={})
    out = mesh_utils.remesh_faces(tmp,face_ids,{})
    cap_model.set_surface(out)
    all_faces = cap_model.get_face_ids()
    faces = cap_model.identify_caps()
    face_ids = [all_faces[i] for i,x in enumerate(faces) if x]
    for cap in face_ids:
        face_edge_size = {}
        cap_model.set_surface(out)
        face_elements = cap_model.get_face_polydata(cap).GetNumberOfCells()
        if face_elements < n_cells:
            print('Face Edge Size: {{}} ---> {{}}'.format(face_edge_size,round(face_edge_size*(face_elements/n_cells),5)))
            out = mesh_utils.remesh_faces(out,[cap],round(face_edge_size*(face_elements/n_cells),5))
    model = modeling.PolyData()
    model.set_surface(out)
    better.append(model.get_polydata())
##############################
# INITIAL CLEAN BAD EDGES
##############################
fe = vtk.vtkFeatureEdges()
fe.FeatureEdgesOff()
fe.BoundaryEdgesOn()
fe.NonManifoldEdgesOn()
clean = vtk.vtkCleanPolyData()
clean2 = vtk.vtkCleanPolyData()
fill = vtk.vtkFillHolesFilter()
tri = vtk.vtkTriangleFilter()
for idx,vessel in enumerate(better):
    fe.SetInputDataObject(vessel)
    fe.Update()
    BAD_CELLS = fe.GetOutput().GetNumberOfCells()
    print(BAD_CELLS)
    if BAD_CELLS > 0:
        print('Attempt initial clean')
        clean.SetInputDataObject(vessel)
        clean.Update()
        tmp_vessel = clean.GetOutput()
        fe.SetInputDataObject(tmp_vessel)
        BAD_CELLS = fe.GetOutput().GetNumberOfCells()
        print(BAD_CELLS)
        if BAD_CELLS > 0:
            print('Attempt initial hole filling')
            fill.SetInputDataObject(tmp_vessel)
            fill.Update()
            tri.SetInputDataObject(fill.GetOutput())
            tri.Update()
            clean2.SetInputDataObject(tri.GetOutput())
            clean2.SetConvertLinesToPoints(1)
            clean2.SetConvertPolysToLines(1)
            clean2.SetConvertStripsToPolys(1)
            clean2.SetPointMerging(1)
            tmp_vessel = clean2.GetOutput()
            fe.SetInputDataObject(tmp_vessel)
            BAD_CELLS = fe.GetOutput().GetNumberOfCells()
            print(BAD_CELLS)
        better[idx] = tmp_vessel
##############################
# UNION
##############################
model_1 = modeling.PolyData()
model_1.set_surface(better[0])
model_2 = modeling.PolyData()
model_2.set_surface(better[1])
unioned_model = modeler.union(model_1,model_2)
fill.SetHoleSize(fill.GetHoleSizeMaxValue())
clean.SetInputDataObject(unioned_model.get_polydata())
clean.Update()
#clean2 = vtk.vtkCleanPolyData()
region = vtk.vtkPolyDataConnectivityFilter()
strip = vtk.vtkStripper()
poly = vtk.vtkPolyData()
tmp = clean.GetOutput()
fe.SetInputDataObject(tmp)
fe.Update()
terminating = False
print('Union: {{}}\\nBad Edges: {{}}'.format(1,fe.GetOutput().GetNumberOfCells()))
for i in range(2,len(better)):
    model_1.set_surface(tmp)
    #model_1.compute_boundary_faces(angle=45)
    tmp = mesh_utils.remesh(model_1.get_polydata(),hmin=0.01,hmax=0.02)
    model_1.set_surface(tmp)
    model_2.set_surface(better[i])
    print('models created')
    unioned_model = modeler.union(model_1,model_2)
    print('union successful')
    fe.SetInputDataObject(unioned_model.get_polydata())
    fe.Update()
    print('Initial Bad Edges: {{}}'.format(fe.GetOutput().GetNumberOfCells()))
    clean.SetInputDataObject(unioned_model.get_polydata())
    clean.Update()
    tmp = clean.GetOutput()
    fe.SetInputDataObject(tmp)
    fe.Update()
    print('Union: {{}}\\nBad Edges: {{}}'.format(i,fe.GetOutput().GetNumberOfCells()))
    if fe.GetOutput().GetNumberOfCells() == 0:
        run_remesh = False
    if fe.GetOutput().GetNumberOfCells() > 0:
        number_bad_edges = fe.GetOutput().GetNumberOfLines()
        fill.SetInputDataObject(clean.GetOutput())
        fill.Update()
        #region.SetInputDataObject(fe.GetOutput())
        #region.SetExtractionMode(6)
        #region.Update()
        #strip.SetInputDataObject(region.GetOutput())
        #strip.Update()
        tri.SetInputDataObject(fill.GetOutput())
        tri.Update()
        clean2.SetInputDataObject(tri.GetOutput())
        clean2.SetConvertLinesToPoints(1)
        clean2.SetConvertPolysToLines(1)
        clean2.SetConvertStripsToPolys(1)
        clean2.SetPointMerging(1)
        clean2.Update()
        fe.SetInputDataObject(clean2.GetOutput())
        fe.Update()
        number_bad_edges = fe.GetOutput().GetNumberOfLines()
        tmp = clean2.GetOutput()
        print('Pedantic Cleaning: '+str(number_bad_edges))
        colors = vtk.vtkNamedColors()
        actor = vtk.vtkActor()
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputDataObject(tmp)
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(colors.GetColor3d('red'))
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(colors.GetColor3d('white'))
        render_window = vtk.vtkRenderWindow()
        render_window.AddRenderer(renderer)
        render_window.SetWindowName('Polydata')
        interactor = vtk.vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)
        renderer.AddActor(actor)
        render_window.Render()
        interactor.Start()
        colors = vtk.vtkNamedColors()
        actor = vtk.vtkActor()
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputDataObject(better[i])
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(colors.GetColor3d('blue'))
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(colors.GetColor3d('white'))
        render_window = vtk.vtkRenderWindow()
        render_window.AddRenderer(renderer)
        render_window.SetWindowName('Polydata')
        interactor = vtk.vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)
        renderer.AddActor(actor)
        render_window.Render()
        interactor.Start()
        if number_bad_edges:
            terminating = True
        else:
            terminating = False
    #colors = vtk.vtkNamedColors()
    #actor = vtk.vtkActor()
    #mapper = vtk.vtkPolyDataMapper()
    #mapper.SetInputDataObject(tmp)
    #actor.SetMapper(mapper)
    #actor.GetProperty().SetColor(colors.GetColor3d('red'))
    #renderer = vtk.vtkRenderer()
    #renderer.SetBackground(colors.GetColor3d('white'))
    #render_window = vtk.vtkRenderWindow()
    #render_window.AddRenderer(renderer)
    #render_window.SetWindowName('Polydata')
    #interactor = vtk.vtkRenderWindowInteractor()
    #interactor.SetRenderWindow(render_window)
    #renderer.AddActor(actor)
    #render_window.Render()
    #interactor.Start()
############################
# COMBINE FACES
############################
if not terminating:
    model.set_surface(tmp)
    model.compute_boundary_faces({})
    caps = model.identify_caps()
    ids = model.get_face_ids()
    walls = [ids[i] for i,x in enumerate(caps) if not x]
    while len(walls) > 1:
        target = walls[0]
        lose = walls[1]
        combined = mesh_utils.combine_faces(model.get_polydata(),target,lose)
        model.set_surface(combined)
        ids = model.get_face_ids()
        caps = model.identify_caps()
        walls = [ids[i] for i,x in enumerate(caps) if not x]
        print(walls)
    ids = model.get_face_ids()
    if {}:
        dmg.add_model({},model)
    if len(ids) > len(better)+2:
        face_cells = []
        for idx in ids:
            face = model.get_face_polydata(idx)
            cells = face.GetNumberOfCells()
            print(cells)
            face_cells.append(cells)
        data_to_remove = len(ids) - (len(better)+2)
        remove_list = []
        for i in range(data_to_remove):
            remove_list.append(ids[face_cells.index(min(face_cells))])
            face_cells[face_cells.index(min(face_cells))] += 1000
        print(remove_list)
        while len(remove_list) > 0:
            target = walls[0]
            lose = remove_list.pop(-1)
            combined = mesh_utils.combine_faces(model.get_polydata(),target,lose)
            model.set_surface(combined)
            print(remove_list)
        print(model.get_face_ids())
    ###############################
    # LOCAL SMOOTHING (not included)
    ###############################
    #smoothing_params = {{'method':'constrained', 'num_iterations':5, 'constrain_factor':0.2, 'num_cg_solves':30}}
    smooth_model = model.get_polydata()
    for idx, contour_set in enumerate(contour_list):
         if idx == 0:
              continue
         smoothing_params = {{'method':'constrained', 'num_iterations':3, 'constrain_factor':0.1+(0.9*(1-contour_set[0].get_radius()/contour_list[0][0].get_radius())), 'num_cg_solves':30}}
         smooth_model = geometry.local_sphere_smooth(smooth_model,contour_set[0].get_radius()*2,contour_set[0].get_center(),smoothing_params)
         print('local sphere smoothing {{}}'.format(idx))
    model.set_surface(smooth_model)

"""



import numpy as np
from .kernel_functions.fitness import fitness
from .kernel_functions.grad import grad
from .kernel_functions.hess import hess
#import sys #should remove

#sys.path.append('..')
from ..core.m_matrix import M
from ..core.n_matrix import N
from ..core.a_matrix import A
from ..core.h_matrix import H
from ..core.utils import norm

class kernel:
    def __init__(self,points):
        self.ndim = points.shape[0]
        self.ddim = points.shape[1]
        self.A_inv,self.K00,self.K01,self.K11 = A(points)
        self.H_0 = H(self.K00,self.K01,self.K11,0)
        self.fit  = lambda x: fitness(x,self.H_0)
        self.jac  = lambda x: grad(x,self.H_0,n=points.shape[0])
        self.hess = lambda x: hess(x,self.H_0,n=points.shape[0])
    def fitness(self,x):
        return self.fit(x)
    def get_bounds(self):
        bounds = []
        lb = []
        ub = []
        for i in range(self.ndim):
            for j in range(self.ddim-1):
                if j < self.ddim-2:
                    lb.append(-2*np.pi)
                    ub.append(2*np.pi)
                else:
                    lb.append(-2*np.pi)
                    ub.append(2*np.pi)
        bounds.append(lb)
        bounds.append(ub)
        return tuple(bounds)
    def gradient(self,x):
        return self.jac(x)
    def hessian(self,x):
        return self.hess(x)
